/*#io
GLUT ioDoc(
			 docCopyright("Steve Dekorte", 2002)
			 docLicense("BSD revised")
			 docCategory("Graphics")
 */

#include "IoGLUT.h"

#include "List.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoCFunction.h"
#include "IoSeq.h"
#include "IoList.h"
#include "List.h"
#include "IoVector.h"
#include <time.h>

#define DATA(self) ((IoGLUTData *)self->data)

void IoGlutTimerFunc(int vv);

static IoGLUT *proto = 0x0;


IoTag *IoGLUT_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("GLUT"); 
	tag->state = state;
	tag->cloneFunc = (TagCloneFunc *)IoGLUT_rawClone;
	tag->freeFunc  = (TagFreeFunc *)IoGLUT_free;
	tag->markFunc  = (TagMarkFunc *)IoGLUT_mark;
	return tag;
}

#define GLUTMESSAGE(name) \
IoMessage_newWithName_label_(state, IOSYMBOL(name), IOSYMBOL("[GLUT]"))

IoGLUT *IoGLUT_proto(void *state)
{
	IoObject *self = IoObject_new(state);
	/*printf("state = %p proto = %p self = %p\n", state, proto, self); */
	
	proto = self;
	self->tag = IoGLUT_tag(state);
	self->data = calloc(1, sizeof(IoGLUTData));
	
	DATA(self)->eventTarget = 0x0;
	DATA(self)->entryMessage = GLUTMESSAGE("entry");
	DATA(self)->displayMessage = GLUTMESSAGE("display");
	DATA(self)->keyboardMessage = GLUTMESSAGE("keyboard");
	DATA(self)->keyboardUpMessage = GLUTMESSAGE("keyboardUp");
	DATA(self)->motionMessage = GLUTMESSAGE("motion");
	DATA(self)->passiveMotionMessage = GLUTMESSAGE("passiveMotion");
	DATA(self)->mouseMessage = GLUTMESSAGE("mouse");
	DATA(self)->reshapeMessage = GLUTMESSAGE("reshape");
	DATA(self)->specialMessage = GLUTMESSAGE("special");
	DATA(self)->timerMessage = GLUTMESSAGE("timer");
	
	DATA(self)->dragMessage = GLUTMESSAGE("drag");
	DATA(self)->dropMessage = GLUTMESSAGE("drop");
	DATA(self)->acceptsDropMessage = GLUTMESSAGE("acceptsDrop");
	DATA(self)->copyMessage = GLUTMESSAGE("copy");
	DATA(self)->pasteMessage = GLUTMESSAGE("paste");
	DATA(self)->deleteMessage = GLUTMESSAGE("delete");
	DATA(self)->nanoSleepPeriod = 1000000000/2;
	
	IoState_registerProtoWithFunc_(state, self, IoGLUT_proto);
	
	IoGLUT_protoInit(self);
	return self;
}

IoGLUT *IoGLUT_new(void *state)
{ 
	return IoState_protoWithInitFunction_(state, IoGLUT_proto); 
}

void IoGLUT_free(IoGLUT *self) 
{ 
	/* add code to shut down GLUT */
	free(self->data); 
}

void IoGLUT_mark(IoGLUT *self)
{
	if (DATA(self)->eventTarget) 
	{
		IoObject_makeGrayIfWhite(DATA(self)->eventTarget);
	}
	
	IoObject_makeGrayIfWhite(DATA(self)->entryMessage);
	IoObject_makeGrayIfWhite(DATA(self)->displayMessage);
	IoObject_makeGrayIfWhite(DATA(self)->keyboardMessage);
	IoObject_makeGrayIfWhite(DATA(self)->keyboardUpMessage);
	IoObject_makeGrayIfWhite(DATA(self)->motionMessage);
	IoObject_makeGrayIfWhite(DATA(self)->passiveMotionMessage);
	IoObject_makeGrayIfWhite(DATA(self)->mouseMessage);
	IoObject_makeGrayIfWhite(DATA(self)->reshapeMessage);
	IoObject_makeGrayIfWhite(DATA(self)->specialMessage);
	IoObject_makeGrayIfWhite(DATA(self)->timerMessage);
	
	IoObject_makeGrayIfWhite(DATA(self)->dragMessage);
	IoObject_makeGrayIfWhite(DATA(self)->acceptsDropMessage);
	IoObject_makeGrayIfWhite(DATA(self)->dropMessage);
	IoObject_makeGrayIfWhite(DATA(self)->copyMessage);
	IoObject_makeGrayIfWhite(DATA(self)->pasteMessage);
	IoObject_makeGrayIfWhite(DATA(self)->deleteMessage);
}

/* ----------------------------------------------------------- */

IoObject *IoGLUT_rawClone(IoGLUT *self)
{ 
	return IoState_protoWithInitFunction_(IOSTATE, IoGLUT_proto); 
}

/* ----------------------------------------------------------- */

IoObject *IoGLUT_glutInitDisplayMode(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int mode = IoMessage_locals_intArgAt_(m, locals, 0);
	glutInitDisplayMode(mode);
	return self;
}

IoObject *IoGLUT_glutInitWindowSize(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	int width  = IoMessage_locals_intArgAt_(m, locals, 0);
	int height = IoMessage_locals_intArgAt_(m, locals, 1);
	glutInitWindowSize(width, height);
	return self; 
}

IoObject *IoGLUT_glutInitWindowPosition(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	int x = IoMessage_locals_intArgAt_(m, locals, 0);
	int y = IoMessage_locals_intArgAt_(m, locals, 1);
	glutInitWindowPosition(x, y);
	return self; 
}

void IoGlutIdleFunc(void) 
{ 
	IoState_yield(proto->tag->state); 
}

IoObject *IoGLUT_glutInit(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	IoState *state = IOSTATE;
	glutInit(&(state->mainArgs->argc), (char **)(state->mainArgs->argv));
    //glutIdleFunc(IoGlutIdleFunc);
    //glutTimerFunc((unsigned int)10, IoGlutTimerFunc, -1); 
	return self;
}

IoObject *IoGLUT_glutInitDisplayString(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	glutInitDisplayString(IoSeq_asCString(s));
	return self;
}

IoObject *IoGLUT_glutCreateWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	int windowId = glutCreateWindow(IoSeq_asCString(s));
	return IONUMBER(windowId);
}

IoObject *IoGLUT_glutCreateSubWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int win = IoMessage_locals_intArgAt_(m, locals, 0);
	int x = IoMessage_locals_intArgAt_(m, locals, 1);
	int y = IoMessage_locals_intArgAt_(m, locals, 2);
	int w = IoMessage_locals_intArgAt_(m, locals, 3);
	int h = IoMessage_locals_intArgAt_(m, locals, 4);
	return IONUMBER(glutCreateSubWindow(win, x, y, w, h));
}

IoObject *IoGLUT_glutSetWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int win = IoMessage_locals_intArgAt_(m, locals, 0);
	glutSetWindow(win);
	return self; 
}

IoObject *IoGLUT_glutGetWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	return IONUMBER(glutGetWindow()); 
}

IoObject *IoGLUT_glutDestroyWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int win = IoMessage_locals_intArgAt_(m, locals, 0);
	glutDestroyWindow(win);
	return self; 
}

IoObject *IoGLUT_glutPositionWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int x = IoMessage_locals_intArgAt_(m, locals, 0);
	int y = IoMessage_locals_intArgAt_(m, locals, 1);
	glutPositionWindow(x, y);
	return self;
}

IoObject *IoGLUT_glutReshapeWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int w = IoMessage_locals_intArgAt_(m, locals, 0);
	int h = IoMessage_locals_intArgAt_(m, locals, 1);
	glutReshapeWindow(w, h);
	return self;
}

IoObject *IoGLUT_glutPopWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutPopWindow(); return self; 
}

IoObject *IoGLUT_glutPushWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutPushWindow(); 
	return self; 
}

IoObject *IoGLUT_glutShowWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutShowWindow(); 
	return self; 
}

IoObject *IoGLUT_glutHideWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutHideWindow(); 
	return self; 
}

IoObject *IoGLUT_glutIconifyWindow(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutIconifyWindow(); 
	return self; 
}

IoObject *IoGLUT_glutSetWindowTitle(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	glutSetWindowTitle(CSTRING(s));
	return self;
}

IoObject *IoGLUT_glutSetIconTitle(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	glutSetIconTitle(CSTRING(s));
	return self;
}  

IoObject *IoGLUT_glutEventTarget_(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	DATA(proto)->eventTarget = IOREF(IoMessage_locals_valueArgAt_(m, locals, 0));
	return self;
}

/* --- events --- */

void IoGlutDisplayFunc(void)
{
	IoState_pushRetainPool(proto->tag->state);
	
	if (DATA(proto)->eventTarget)
	{ 
		IoState_withinCatchPerform(proto->tag->state, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->displayMessage); 
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutDisplayFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutDisplayFunc(IoGlutDisplayFunc); 
	return self; 
}

void IoGlutKeyboardFunc(unsigned char key, int xv, int yv)
{
	IoState_pushRetainPool(proto-> tag->state);
	IoMessage_setCachedArg_toInt_(DATA(proto)->keyboardMessage, 0, (int)key);
	IoMessage_setCachedArg_toInt_(DATA(proto)->keyboardMessage, 1, xv);
	IoMessage_setCachedArg_toInt_(DATA(proto)->keyboardMessage, 2, yv);
	
	if (DATA(proto)->eventTarget)
	{ 
		IoState_withinCatchPerform(proto->tag->state, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->keyboardMessage); 
	}
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutKeyboardFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutKeyboardFunc(IoGlutKeyboardFunc); 
	return self; 
}

void IoGlutSpecialFunc(int key, int xv, int yv)
{
	IoState_pushRetainPool(proto-> tag->state);
	IoMessage_setCachedArg_toInt_(DATA(proto)->specialMessage, 0, (int)key);
	IoMessage_setCachedArg_toInt_(DATA(proto)->specialMessage, 1, xv);
	IoMessage_setCachedArg_toInt_(DATA(proto)->specialMessage, 2, yv);
	
	if (DATA(proto)->eventTarget)
	{ 
		IoState_withinCatchPerform(proto->tag->state, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->specialMessage); 
	}
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutSpecialFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutSpecialFunc(IoGlutSpecialFunc); 
	return self; 
}

void IoGlutKeyboardUpFunc(unsigned char key, int xv, int yv)
{
	IoState_pushRetainPool(proto->tag->state);
	IoMessage_setCachedArg_toInt_(DATA(proto)->keyboardUpMessage, 0, (int)key);
	IoMessage_setCachedArg_toInt_(DATA(proto)->keyboardUpMessage, 1, xv);
	IoMessage_setCachedArg_toInt_(DATA(proto)->keyboardUpMessage, 2, yv);
	
	if (DATA(proto)->eventTarget)
	{ 
		IoState_withinCatchPerform(proto->tag->state, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->keyboardUpMessage); 
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutKeyboardUpFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
#if (GLUT_API_VERSION >= 4 || GLUT_XLIB_IMPLEMENTATION >= 13)
	glutKeyboardUpFunc(IoGlutKeyboardUpFunc); 
	return self; 
#endif
}

IoObject *IoGLUT_glutGetModifiers(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	return IONUMBER(glutGetModifiers());
}

void IoGlutEntryFunc(int state)
{
	IoState_pushRetainPool(proto->tag->state);
	IoMessage_setCachedArg_toInt_(DATA(proto)->entryMessage, 0, state);
	
	if (DATA(proto)->eventTarget)
	{ 
		IoState_withinCatchPerform(proto->tag->state, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->entryMessage); 
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutEntryFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutEntryFunc(IoGlutEntryFunc); return self; 
}

void IoGlutMotionFunc(int xv, int yv)
{
	IoState_pushRetainPool(proto->tag->state);
	
	{
		IoMessage_setCachedArg_toInt_(DATA(proto)->motionMessage, 0, xv);
		IoMessage_setCachedArg_toInt_(DATA(proto)->motionMessage, 1, yv);
		
		if (DATA(proto)->eventTarget) 
		{ 
			IoState_withinCatchPerform(proto->tag->state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->motionMessage); 
		}
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutMotionFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutMotionFunc(IoGlutMotionFunc); 
	return self; 
}

void IoGlutPassiveMotionFunc(int xv, int yv)
{
	IoState_pushRetainPool(proto->tag->state);
	
	{
		IoMessage_setCachedArg_toInt_(DATA(proto)->passiveMotionMessage, 0, xv);
		IoMessage_setCachedArg_toInt_(DATA(proto)->passiveMotionMessage, 1, yv);
		
		if (DATA(proto)->eventTarget)
		{ 
			IoState_withinCatchPerform(proto->tag->state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->passiveMotionMessage); 
		}
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutPassiveMotionFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutPassiveMotionFunc(IoGlutPassiveMotionFunc); 
	return self; 
}

void IoGlutMouseFunc(int button, int state, int xv, int yv) 
{
	IoState_pushRetainPool(proto->tag->state);
	
	{
		IoMessage_setCachedArg_toInt_(DATA(proto)->mouseMessage, 0, button);
		IoMessage_setCachedArg_toInt_(DATA(proto)->mouseMessage, 1, state);
		IoMessage_setCachedArg_toInt_(DATA(proto)->mouseMessage, 2, xv);
		IoMessage_setCachedArg_toInt_(DATA(proto)->mouseMessage, 3, yv);
		
		if (DATA(proto)->eventTarget) 
		{ 
			IoState_withinCatchPerform(proto->tag->state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->mouseMessage); 
		}
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutMouseFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutMouseFunc(IoGlutMouseFunc); 
	return self; 
}

void IoGlutReshapeFunc(int width, int height)
{
	IoState_pushRetainPool(proto->tag->state);
	{
		IoMessage_setCachedArg_toInt_(DATA(proto)->reshapeMessage, 0, width?width:1);
		IoMessage_setCachedArg_toInt_(DATA(proto)->reshapeMessage, 1, height?height:1);
		
		if (DATA(proto)->eventTarget) 
		{ 
			IoState_withinCatchPerform(proto->tag->state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->reshapeMessage); 
		}
	}
	
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutReshapeFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutReshapeFunc(IoGlutReshapeFunc); 
	return self; 
}

void IoGlutTimerFunc(int vv)
{
	IoState_pushRetainPool(proto->tag->state);
	
	if (vv == -1)
	{
		IoState_yield(proto->tag->state); 
		glutTimerFunc((unsigned int)100, IoGlutTimerFunc, -1); 
	}
	else
	{
		IoMessage_setCachedArg_toInt_(DATA(proto)->timerMessage, 0, vv);
		
		if (DATA(proto)->eventTarget)
		{ 
			IoState_withinCatchPerform(proto->tag->state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->timerMessage); 
		}
	}
	IoState_popRetainPool(proto->tag->state);
}

IoObject *IoGLUT_glutTimerFunc(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	int msecs = IoMessage_locals_intArgAt_(m, locals, 0);
	int v = IoMessage_locals_intArgAt_(m, locals, 1);
	glutTimerFunc((unsigned int)msecs, IoGlutTimerFunc, v); 
	return self; 
}

#ifdef GLUT_KEY_REPEAT_ON
IoObject *IoGLUT_glutIgnoreKeyRepeat(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	int v = IoMessage_locals_intArgAt_(m, locals, 0);
	glutIgnoreKeyRepeat(v); 
	return self; 
}
#endif

IoObject *IoGLUT_glutMainLoop(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutMainLoop(); 
	return self; 
}

IoObject *IoGLUT_glutPostRedisplay(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutPostRedisplay(); 
	return self; 
}

IoObject *IoGLUT_glutSwapBuffers(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutSwapBuffers(); 
	return self; 
}

IoObject *IoGLUT_glutSolidCone(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble radius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	GLdouble height = IoMessage_locals_intArgAt_(m, locals, 1);
	GLint slices = IoMessage_locals_intArgAt_(m, locals, 2);
	GLint stacks = IoMessage_locals_intArgAt_(m, locals, 3);
	glutSolidCone(radius, height, slices, stacks);
	return self;
}

IoObject *IoGLUT_glutWireCone(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble radius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	GLdouble height = IoMessage_locals_intArgAt_(m, locals, 1);
	GLint slices = IoMessage_locals_intArgAt_(m, locals, 2);
	GLint stacks = IoMessage_locals_intArgAt_(m, locals, 3);
	glutWireCone(radius, height, slices, stacks);
	return self;
}

IoObject *IoGLUT_glutSolidSphere(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble radius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	GLint slices = IoMessage_locals_intArgAt_(m, locals, 1);
	GLint stacks = IoMessage_locals_intArgAt_(m, locals, 2);
	glutSolidSphere(radius, slices, stacks);
	return self;
}

IoObject *IoGLUT_glutWireSphere(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble radius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	GLint slices = IoMessage_locals_intArgAt_(m, locals, 1);
	GLint stacks = IoMessage_locals_intArgAt_(m, locals, 2);
	glutWireSphere(radius, slices, stacks);
	return self;
}

IoObject *IoGLUT_glutSolidTorus(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble innerRadius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	GLdouble outerRadius = IoMessage_locals_doubleArgAt_(m, locals, 1);
	GLint nsides = IoMessage_locals_intArgAt_(m, locals, 2);
	GLint rings = IoMessage_locals_intArgAt_(m, locals, 3);
	glutSolidTorus(innerRadius, outerRadius, nsides, rings);
	return self;
}

IoObject *IoGLUT_glutWireTorus(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble innerRadius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	GLdouble outerRadius = IoMessage_locals_doubleArgAt_(m, locals, 1);
	GLint nsides = IoMessage_locals_intArgAt_(m, locals, 2);
	GLint rings = IoMessage_locals_intArgAt_(m, locals, 3);
	glutWireTorus(innerRadius, outerRadius, nsides, rings);
	return self;
}

IoObject *IoGLUT_glutSolidDodecahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutSolidDodecahedron(); 
	return self; 
}

IoObject *IoGLUT_glutWireDodecahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutWireDodecahedron(); 
	return self; 
}

IoObject *IoGLUT_glutSolidOctahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutSolidOctahedron(); 
	return self; 
}

IoObject *IoGLUT_glutWireOctahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutWireOctahedron(); 
	return self; 
}

IoObject *IoGLUT_glutSolidTetrahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutSolidTetrahedron(); 
	return self; 
}

IoObject *IoGLUT_glutWireTetrahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutWireTetrahedron(); 
	return self; 
}

IoObject *IoGLUT_glutSolidIcosahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutSolidIcosahedron(); 
	return self; 
}

IoObject *IoGLUT_glutWireIcosahedron(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	glutWireIcosahedron(); 
	return self; 
}

IoObject *IoGLUT_glutSolidTeapot(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble r = IoMessage_locals_doubleArgAt_(m, locals, 0);
	glutSolidTeapot(r); 
	return self; 
}

IoObject *IoGLUT_glutWireTeapot(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	GLdouble r = IoMessage_locals_doubleArgAt_(m, locals, 0);
	glutWireTeapot(r); 
	return self; 
}

IoObject *IoGLUT_glutSolidCube(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	GLdouble radius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	glutSolidCube(radius);
	return self;
}

IoObject *IoGLUT_glutWireCube(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	GLdouble radius = IoMessage_locals_doubleArgAt_(m, locals, 0);
	glutWireCube(radius);
	return self;
}

IoObject *IoGLUT_glutStrokeCharacter(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	int fontNum = IoMessage_locals_intArgAt_(m, locals, 0);
	char c = IoMessage_locals_intArgAt_(m, locals, 1);
	void *font = GLUT_STROKE_ROMAN;
	if (fontNum) font = GLUT_STROKE_MONO_ROMAN;
	glutStrokeCharacter(font, c);
	return self;
}

IoObject *IoGLUT_glutStrokeString(IoGLUT *self, IoObject *locals, IoMessage *m)
{ 
	int fontNum = IoMessage_locals_intArgAt_(m, locals, 0);
	IoSymbol *string = IoMessage_locals_seqArgAt_(m, locals, 1);
	char *s = CSTRING(string);
	void *font = GLUT_STROKE_ROMAN;
	if (fontNum) font = GLUT_STROKE_MONO_ROMAN;
	
	while (*s) 
	{
		glutStrokeCharacter(font, *s);
		s++;
	}
	
	return self;
}

/* --- GLUT game mode ----------------------------------------------------- */

#ifdef GLUT_GAME_MODE_ACTIVE

IoObject *IoGLUT_glutEnterGameMode(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	glutEnterGameMode();
	return self;
}

IoObject *IoGLUT_glutLeaveGameMode(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	glutLeaveGameMode();
	return self;
}

IoObject *IoGLUT_glutWarpPointer(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int x = IoMessage_locals_intArgAt_(m, locals, 0);
	int y = IoMessage_locals_intArgAt_(m, locals, 1);
	glutWarpPointer(x, y);
	return self;
}

IoObject *IoGLUT_glutGameModeString(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	glutGameModeString(IoSeq_asCString(s));
	return self;
}

#endif

IoObject *IoGLUT_glutSetCursor(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int x = IoMessage_locals_intArgAt_(m, locals, 0);
	glutSetCursor(x);
	return self;
}

/* -------------------------------------------------------------------*/

/* can use NULL tag since the methods don't access DATA(self) - they use DATA(proto) instead */
#define IOCFUNCTION_GLUT(func) IOCFUNCTION(func, NULL)

#include "IoGLUTconst.h"
#include "IoGLUTfunc.h"

void IoGLUT_protoInit(IoGLUT *self) 
{     
	IoObject_setSlot_to_(self,IOSYMBOL("clone"), IOCFUNCTION_GLUT(IoObject_self));
	
	/* GLUT Constants */
	{
		t_ioGLUT_constTable *curpos=ioGLUT_constTable;
		while (curpos->name) 
		{
			IoObject_setSlot_to_(self, 
							 IOSYMBOL(curpos->name), 
							 IONUMBER(curpos->value));
			curpos++;
		}
	}
	
	/* GLUT Functions */
	{
		t_ioGLUT_funcTable *curpos=ioGLUT_funcTable;
		while (curpos->name) 
		{
			IoObject_setSlot_to_(self, 
							 IOSYMBOL(curpos->name), 
							 IOCFUNCTION_GLUT(curpos->func));
			curpos++;
		}
	}
}

// --- Extra In --------------------------------------------------------------

int IoGlutAcceptsDropFunc(
					 int x, 
					 int y, 
					 const char *type, 
					 const unsigned char *data, 
					 int dataLength)
{
	int result = 0;
	IoState *state = proto->tag->state;
	IoState_pushRetainPool(state);
	
	{
		IoMessage *m = DATA(proto)->acceptsDropMessage;
		IoSymbol *typeString = IoState_symbolWithCString_(state, (char *)type);
		IoSeq *dataBuffer = IoSeq_newWithData_length_(state, (unsigned char *)data, dataLength);
		
		IoMessage_setCachedArg_toInt_(m, 0, x);
		IoMessage_setCachedArg_toInt_(m, 1, y);
		IoMessage_setCachedArg_to_(m, 2, typeString);
		IoMessage_setCachedArg_to_(m, 3, dataBuffer);
		
		if (DATA(proto)->eventTarget) 
		{ 
			IoObject *r = IoState_withinCatchPerform(state, 
											 DATA(proto)->eventTarget, 
											 DATA(proto)->eventTarget, 
											 m); 
			if (r && ISNUMBER(r)) 
			{
				result = CNUMBER(r);
			}
		}
	}
	
	IoState_popRetainPool(state);
	return result;
}

void IoGlutDropFunc(
				int x, 
				int y, 
				const char *type, 
				const unsigned char *data, 
				int dataLength)
{
	IoState *state = proto->tag->state;
	IoState_pushRetainPool(state);
	
	{
		IoMessage *m = DATA(proto)->dropMessage;
		IoSymbol *typeString = IoState_symbolWithCString_(state, (char *)type);
		IoSeq *dataBuffer = IoSeq_newWithData_length_(state, (unsigned char *)data, dataLength);
		
		IoMessage_setCachedArg_toInt_(m, 0, x);
		IoMessage_setCachedArg_toInt_(m, 1, y);
		IoMessage_setCachedArg_to_(m, 2, typeString);
		IoMessage_setCachedArg_to_(m, 3, dataBuffer);
		
		if (DATA(proto)->eventTarget) 
		{ 
			IoState_withinCatchPerform(proto->tag->state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  m); 
		}
	}
	
	IoState_popRetainPool(state);
}

void IoGlutPasteFunc(
				 const char *type, 
				 const unsigned char *data, 
				 int dataLength) 
{
	IoState *state = proto->tag->state;
	IoState_pushRetainPool(state);
	
	{
		IoMessage *m = DATA(proto)->pasteMessage;
		IoSymbol *typeString = IoState_symbolWithCString_(state, (char *)type);
		IoSeq *dataBuffer = IoSeq_newWithData_length_(state, (unsigned char *)data, dataLength);
		
		IoMessage_setCachedArg_to_(m, 0, typeString);
		IoMessage_setCachedArg_to_(m, 1, dataBuffer);
		
		if (DATA(proto)->eventTarget) 
		{ 
			IoState_withinCatchPerform(state, 
								  DATA(proto)->eventTarget, 
								  DATA(proto)->eventTarget, 
								  m); 
		}
	}
	
	IoState_popRetainPool(proto->tag->state);
}

void IoGlutDeleteFunc(void)
{
	IoState *state = proto->tag->state;
	IoState_pushRetainPool(state);
	if (DATA(proto)->eventTarget) 
	{ 
		IoMessage *m = DATA(proto)->deleteMessage;
		IoState_withinCatchPerform(proto->tag->state, 
							  DATA(proto)->eventTarget, 
							  DATA(proto)->eventTarget, 
							  m); 
	}
	IoState_popRetainPool(state); 
}


// --- Extra Out --------------------------------------------------------------

void IoGLUT_setDragCallback_(IoGLUT *self, DragCallback *func)
{ 
	DATA(self)->dragCallback = func; 
}

IoObject *IoGLUT_drag(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	int x = IoMessage_locals_intArgAt_(m, locals, 0);
	int y = IoMessage_locals_intArgAt_(m, locals, 1);
	IoSymbol *type = IoMessage_locals_symbolArgAt_(m, locals, 2);
	IoSeq *buffer = IoMessage_locals_mutableSeqArgAt_(m, locals, 3);
	
	if (DATA(self)->dragCallback)
	{
		(DATA(self)->dragCallback)(DATA(self)->callbackContext, 
							  x, 
							  y, 
							  CSTRING(type), 
							  IoSeq_rawBytes(buffer), 
							  IoSeq_rawSize(buffer));
	}
	
	return IONIL(self);
}

void IoGLUT_setCopyCallback_(IoGLUT *self, CopyCallback *func)
{ 
	DATA(self)->copyCallback = func; 
}

IoObject *IoGLUT_copy(IoGLUT *self, IoObject *locals, IoMessage *m)
{
	IoSymbol *type = IoMessage_locals_symbolArgAt_(m, locals, 0);
	IoSeq *buffer = IoMessage_locals_mutableSeqArgAt_(m, locals, 1);
	
	if (DATA(self)->copyCallback)
	{
		(DATA(self)->copyCallback)(DATA(self)->callbackContext, 
							  CSTRING(type), 
							  IoSeq_rawBytes(buffer), 
							  IoSeq_rawSize(buffer));
	}
	
	return IONIL(self);
}

