

Sequence writeToStream := method(b, b appendSeq(self))

SGMLTag := Object clone do(
    type := "SGMLTag"
	docCopyright("Steve Dekorte, 2004. All rights reserved.")
	docLicense("See _BSDLicense.txt.")
	docCategory("XML")
	
	docDescription("Representation of an SGML / HTML / XML tag.")
	docSlot("setName(aString)", "Sets the tag name. Returns self.")
	docSlot("name", "Returns the tag name")
	docSlot("attributes", "Returns a Map containing the tag's attributes.")
	docSlot("subitems", "Returns a List containing the tag's subitems.")
	docSlot("asString", "Returns a String representation of the tag and all of it's subitems.")
		
    newSlot("name", nil)
    newSlot("attributes", Map clone)
    newSlot("subitems", List clone)
    
    init := method(
		self attributes := attributes clone
		self subitems := subitems clone
    )
    
    asString := method( 
		b := Sequence clone
		writeToStream(b)
		b asString
    )
    
    writeToStream := method(b,
		if (name,
			b appendSeq("<", name)
			attributes keys sort foreach(k,
				v := attributes at(k) 
				b appendSeq(" ", k, "=\"", v, "\"")
			)
			b appendSeq(">")
		)
		subitems foreach(item, item writeToStream(b))
		if (name, b appendSeq("</", name, ">"))
    )
)

SGMLParser do(
    type := "SGMLParser"
	init := method(
		self stack := List clone
		self rootTag := SGMLTag clone
		stack push(rootTag)
    )
    
    top := method(
		if (stack last, stack last, rootTag)
    )
    
    startElement := method(name,
		tag := SGMLTag clone setName(name)
		top subitems append(tag)
		stack push(tag)
    )

    endElement := method(name,
		stack pop
    )
    
    newAttribute := method(k, v,
		top attributes atPut(k, v)
    )

    newText := method(name,
		top subitems append(name)
    )
    
    tagForString := method(aString,
		parse(aString)
		rootTag
    )
)

Sequence do(
	asHTML := method(SGMLParser clone tagForString(self))
	asXML  := method(SGMLParser clone tagForString(self))
	asSGML := method(SGMLParser clone tagForString(self))
)

