#!/bin/env python
#
"""
Generates a new .vcproj file for IoVM, assuming that all *.c and *.h in
"dir_iovm" and "subdirs" are involved in the build.

Instructions:

1) Place the script in the "projects\win32vc7\IoVM\" directory and execute.
  - "IoVM-autogen.vcproj" is created.
   
2) rename "IoVM-autogen.vcproj" to "IoVM.vcproj" to use.

12/5/2005 -- Kevin Edwards
"""

import os, sys, glob

dir_io = os.path.join( "..", "..", ".." )
dir_iovm = os.path.join( dir_io, "vm" )
subdirs = ["base","SkipDB"]
#output = sys.stdout
output = open("IoVM-autogen.vcproj","w")

iovm_vcproj_prefix = r"""<?xml version="1.0" encoding="Windows-1252"?>
<VisualStudioProject
  ProjectType="Visual C++"
  Version="7.10"
  Name="IoVM"
  ProjectGUID="{D8EA8C47-D814-4F06-B790-DC6344D79D6C}"
  RootNamespace="IoVM"
  Keyword="Win32Proj">
  <Platforms>
    <Platform
      Name="Win32"/>
  </Platforms>
  <Configurations>
    <Configuration
      Name="Debug|Win32"
      OutputDirectory="Debug"
      IntermediateDirectory="Debug"
      ConfigurationType="1"
      CharacterSet="2">
      <Tool
        Name="VCCLCompilerTool"
        Optimization="0"
        AdditionalIncludeDirectories=".\;..\;%(dir_iovm)s;%(dir_iovm)s\base;"
        PreprocessorDefinitions="WIN32;_DEBUG;_CONSOLE;HAS_FIBERS;FREEGLUT_STATIC"
        IgnoreStandardIncludePath="FALSE"
        MinimalRebuild="TRUE"
        BasicRuntimeChecks="3"
        RuntimeLibrary="3"
        UsePrecompiledHeader="0"
        WarningLevel="3"
        Detect64BitPortabilityProblems="FALSE"
        DebugInformationFormat="4"/>
      <Tool
        Name="VCCustomBuildTool"/>
      <Tool
        Name="VCLinkerTool"
        AdditionalDependencies="ws2_32.lib"
        OutputFile="$(OutDir)/IoVM.exe"
        LinkIncremental="2"
        GenerateDebugInformation="TRUE"
        ProgramDatabaseFile="$(OutDir)/IoVM.pdb"
        SubSystem="1"
        TargetMachine="1"/>
      <Tool
        Name="VCMIDLTool"/>
      <Tool
        Name="VCPostBuildEventTool"/>
      <Tool
        Name="VCPreBuildEventTool"/>
      <Tool
        Name="VCPreLinkEventTool"/>
      <Tool
        Name="VCResourceCompilerTool"/>
      <Tool
        Name="VCWebServiceProxyGeneratorTool"/>
      <Tool
        Name="VCXMLDataGeneratorTool"/>
      <Tool
        Name="VCWebDeploymentTool"/>
      <Tool
        Name="VCManagedWrapperGeneratorTool"/>
      <Tool
        Name="VCAuxiliaryManagedWrapperGeneratorTool"/>
    </Configuration>
    <Configuration
      Name="Release|Win32"
      OutputDirectory="Release"
      IntermediateDirectory="Release"
      ConfigurationType="1"
      CharacterSet="2">
      <Tool
        Name="VCCLCompilerTool"
        AdditionalIncludeDirectories=".\;..\;%(dir_iovm)s;%(dir_iovm)s\base;"
        PreprocessorDefinitions="WIN32;NDEBUG;_CONSOLE;HAS_FIBERS;FREEGLUT_STATIC"
        RuntimeLibrary="4"
        UsePrecompiledHeader="0"
        WarningLevel="3"
        Detect64BitPortabilityProblems="FALSE"
        DebugInformationFormat="3"/>
      <Tool
        Name="VCCustomBuildTool"/>
      <Tool
        Name="VCLinkerTool"
        AdditionalDependencies="ws2_32.lib"
        OutputFile="$(OutDir)/IoVM.exe"
        LinkIncremental="1"
        GenerateDebugInformation="TRUE"
        SubSystem="1"
        OptimizeReferences="2"
        EnableCOMDATFolding="2"
        TargetMachine="1"/>
      <Tool
        Name="VCMIDLTool"/>
      <Tool
        Name="VCPostBuildEventTool"/>
      <Tool
        Name="VCPreBuildEventTool"/>
      <Tool
        Name="VCPreLinkEventTool"/>
      <Tool
        Name="VCResourceCompilerTool"/>
      <Tool
        Name="VCWebServiceProxyGeneratorTool"/>
      <Tool
        Name="VCXMLDataGeneratorTool"/>
      <Tool
        Name="VCWebDeploymentTool"/>
      <Tool
        Name="VCManagedWrapperGeneratorTool"/>
      <Tool
        Name="VCAuxiliaryManagedWrapperGeneratorTool"/>
    </Configuration>
  </Configurations>
  <References>
  </References>
""" % { 'dir_iovm': dir_iovm.replace("/","\\") }

iovm_vcproj_postfix = r"""
  <Globals>
  </Globals>
</VisualStudioProject>
"""

def vcproj_subfilter(name, selection):
  lines = [ '\n  <Filter Name="%s" Filter="">' % (name) ]
  
  files = glob.glob(os.path.join(dir_iovm, name, selection))
  for f in files:
    f = f.replace("/","\\")
    lines.append( """<File RelativePath="%s"></File>""" % (f) )
  lines.append('  </Filter>\n')
  return lines


# Files Section
# --------------------------------------
lines = ["<Files>"]

# Source Files
# --------------------------------------
lines.append(
"""
<Filter
  Name="Source Files"
  Filter="cpp;c;cxx;def;odl;idl;hpj;bat;asm;asmx"
  UniqueIdentifier="{4FC737F1-C7A5-4376-A066-2A32D752A2FF}">
"""
)

files = glob.glob( os.path.join(dir_iovm, "*.c") )
for f in files:
  lines.append( '<File RelativePath="%s"></File>' % (f) )

# Subdirectories
for subdir in subdirs:
  lines += vcproj_subfilter(subdir,"*.c")

lines.append("</Filter>")


# Header Files
# --------------------------------------
lines.append(
"""
<Filter
  Name="Header Files"
  Filter="h;hpp;hxx;hm;inl;inc;xsd"
  UniqueIdentifier="{93995380-89BD-4b04-88EB-625FBE52EBFB}">
"""
)

files = glob.glob( os.path.join(dir_iovm, "*.h") )
for f in files:
  lines.append( '<File RelativePath="%s"></File>' % (f) )

# Subdirectories
for subdir in subdirs:
  lines += vcproj_subfilter(subdir,"*.h")

lines.append("</Filter>")

lines.append("</Files>")


# Output
# --------------------------------------
output.write(iovm_vcproj_prefix)
output.write("\n")
output.write("\n".join(lines))
output.write("\n")
output.write(iovm_vcproj_postfix)

output.close()

