// ResourceManager
// Example use:
//  
//  ResourceManager
//  FontManager addPath("Library/Fonts")
//  ImageManager addPath("Library/Themes/Milk")
//
//  systemFont := FontManager item("FreeTypeSans")

ResourceManager := Object clone do(
    protoName := "ResourceManager"
    init := method(
      self paths := Map clone
      self cache := Map clone
      self suffixes := List clone
    )
    resourceProto := Nil

    addPath := method(path,
      paths atPut(path, Directory clone setPath(path))  
    )

    addRelativePath := method(path, addPath(Path with(launchPath, path)))

    removePath := method(path, paths remove(path))

    addSuffix := method(path, suffixes append(path))
    removeSuffix := method(path, suffixes remove(path))

    item := method(itemName, 
      result := itemOrNil(itemName)
      if (result, return result)
      Exception raise(self name, "unable to find resource '" .. itemName .. "'")
    )

    itemOrNil := method(itemName, 
      //write("looking for ", itemName, "\n")
      result := cache at(itemName)
      //if (result , write("found cached ", itemName, "\n"))
      result = find(itemName)
      if (result, return result)
      Nil
    )

    find := method(name, 
      file := fileFor(name)
      if (file,
	resource := resourceProto clone open(file path)
	cache atPut(name, resource)
	//write("found resource ", name, "\n")
	return resource
      )
      Nil
    )

    fileFor := method(name, 
      paths foreach(path, dir,
	suffixes foreach(i, suffix, 
	  fullName := name .. "." .. suffix
	  result := dir at(fullName)
	  //write(self name, " searching: ", Path with(dir path, fullName), "\n")
	  if (result, return result)
	)
      )
      Nil
    )
)

FontManager := ResourceManager clone

//FontManager addPath(Path with(launchPath, "../Library/Fonts"))
FontManager protoName := "FontManager"
FontManager name := "FontManager"
FontManager addSuffix("ttf")
FontManager resourceProto := Font

ImageManager := ResourceManager clone
ImageManager protoName := "ImageManager"
//ImageManager addPath(Path with(launchPath, "../Library/Themes/Milk"))
ImageManager name := "ImageManager"
ImageManager addSuffix("png")
ImageManager addSuffix("jpg")
ImageManager addSuffix("tiff")
ImageManager resourceProto := Image


