/*#io
IoActivationContext ioDoc(
            docCopyright("Steve Dekorte", 2002)
            docLicense("BSD revised")
            docDescription("ActivationContext stores slots related to activation.")
		  docCategory("Core")
*/

#include "IoActivationContext.h"
#include "IoState.h"
#include "IoObject.h"


#define DATA(self) ((IoActivationContextData *)self->data)

IoTag *IoActivationContext_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("ActivationContext");
	tag->state = state;
	tag->cloneFunc = (TagCloneFunc *)IoActivationContext_rawClone;
	tag->markFunc  = (TagMarkFunc *)IoActivationContext_mark;
	tag->freeFunc  = (TagFreeFunc *)IoActivationContext_free;
	tag->writeToStoreOnStreamFunc  = (TagWriteToStoreOnStreamFunc *)IoActivationContext_writeToStore_stream_;
	tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoActivationContext_readFromStore_stream_;
	return tag;
}
 
void IoActivationContext_writeToStore_stream_(IoActivationContext *self, IoStore *store, BStream *stream)
{}

void IoActivationContext_readFromStore_stream_(IoActivationContext *self, IoStore *store, BStream *stream)
{}

void IoActivationContext_initSlots(IoActivationContext *self)
{
	IoObject *ioNil = IOSTATE->ioNil;
	DATA(self)->sender      = ioNil;
	DATA(self)->message     = ioNil;
	DATA(self)->slotContext = ioNil;
	DATA(self)->target      = ioNil;
	DATA(self)->activated   = ioNil;
}

IoActivationContext *IoActivationContext_proto(void *vState)
{
	IoState *state = (IoState *)vState;
	
	IoMethodTable methodTable[] = {
	{"sender",      IoActivationContext_sender},
	{"message",     IoActivationContext_message},
	{"slotContext", IoActivationContext_slotContext},
	{"target",      IoActivationContext_target},
	{"activated",   IoActivationContext_activated},
		
	{NULL, NULL},
	};
	
	IoObject *self = IoObject_new(state);
	
	self->data = calloc(1, sizeof(IoActivationContextData));
	self->tag = IoActivationContext_tag(state);
	IoActivationContext_initSlots(self);
	
	IoState_registerProtoWithFunc_((IoState *)state, self, IoActivationContext_proto);
	
	IoObject_addMethodTable_(self, methodTable);
	return self;
}

IoActivationContext *IoActivationContext_rawClone(IoActivationContext *proto)
{
	IoObject *self = IoObject_rawClonePrimitive(proto);
	self->data = cpalloc(proto->data, sizeof(IoActivationContextData));
	IoActivationContext_initSlots(self);
	return self;
}

IoActivationContext *IoActivationContext_new(IoState *state)
{
	IoObject *proto = IoState_protoWithInitFunction_((IoState *)state, IoActivationContext_proto);
	return IOCLONE(proto);
}

IoActivationContext *IoActivationContext_with(void *state, 
									 IoObject *sender,
									 IoObject *target,
									 IoObject *message,
									 IoObject *slotContext,
									 IoObject *activated)
{
	IoActivationContext *self = IoActivationContext_new(state);
	DATA(self)->sender      = IOREF(sender);
	DATA(self)->target      = IOREF(target);
	DATA(self)->message     = IOREF(message);
	DATA(self)->slotContext = IOREF(slotContext);	
	DATA(self)->activated   = IOREF(activated);
	return self;
}

void IoActivationContext_mark(IoActivationContext *self)
{ 
	IoActivationContextData *d = DATA(self);

	if (d->sender) IoObject_makeGrayIfWhite(d->sender);
	if (d->target) IoObject_makeGrayIfWhite(d->target);
	if (d->message) IoObject_makeGrayIfWhite(d->message);
	if (d->slotContext) IoObject_makeGrayIfWhite(d->slotContext);
	if (d->activated) IoObject_makeGrayIfWhite(d->activated);
}

void IoActivationContext_free(IoActivationContext *self)
{
	free(self->data); 
	self->data = NULL;
}

IoObject *IoActivationContext_sender(IoObject *self, IoObject *locals, IoMessage *m)
{
	return DATA(self)->sender;
}

IoObject *IoActivationContext_message(IoObject *self, IoObject *locals, IoMessage *m)
{
	return DATA(self)->message;
}

IoObject *IoActivationContext_target(IoObject *self, IoObject *locals, IoMessage *m)
{
	return DATA(self)->target;
}

IoObject *IoActivationContext_slotContext(IoObject *self, IoObject *locals, IoMessage *m)
{
	return DATA(self)->slotContext;
}

IoObject *IoActivationContext_activated(IoObject *self, IoObject *locals, IoMessage *m)
{
	return DATA(self)->activated;
}
