/*#io
Map ioDoc(
          docCopyright("Steve Dekorte", 2002)
          docLicense("BSD revised")    
          docDescription("Object wrapper for an Io coroutine.")
		docCategory("Core")
         */

#include "IoCoroutine.h"
#include "IoObject.h"
#include "IoState.h"
#include "IoCFunction.h"
#include "IoSeq.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoList.h"
#include "IoBlock.h"
#include "IoBlock.h"

//#define DEBUG

#define DATA(self) ((IoCoroutineData *)self->data)

IoCoroutine *IoMessage_locals_coroutineArgAt_(IoMessage *self, IoObject *locals, int n)
{
	IoMessage *m = self;
	IoObject *v = IoMessage_locals_valueArgAt_(self, locals, n);
	IOASSERT(ISCOROUTINE(v), "requires coroutine argument");
	return v;
}

void IoCoroutine_writeToStore_stream_(IoCoroutine *self, IoStore *store, BStream *stream)
{
	//IoCoroutineData *data = DATA(self);
}

void IoCoroutine_readFromStore_stream_(IoCoroutine *self, IoStore *store, BStream *stream)
{
	//IoCoroutineData *data = DATA(self);
}

IoTag *IoCoroutine_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("Coroutine");
	tag->state = state;
	tag->freeFunc  = (TagFreeFunc *)IoCoroutine_free;
	tag->cloneFunc = (TagCloneFunc *)IoCoroutine_rawClone;
	tag->markFunc  = (TagMarkFunc *)IoCoroutine_mark;
	tag->writeToStoreOnStreamFunc  = (TagWriteToStoreOnStreamFunc *)IoCoroutine_writeToStore_stream_;
	tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoCoroutine_readFromStore_stream_;
	return tag;
}

IoCoroutine *IoCoroutine_proto(void *state)
{	
	IoObject *self = IoObject_new(state);
	
	self->tag = IoCoroutine_tag(state);
	self->data = calloc(1, sizeof(IoCoroutineData));
	DATA(self)->ioStack = Stack_new();
	
	IoState_registerProtoWithFunc_((IoState *)state, self, IoCoroutine_proto);
	
	DATA(self)->cid = Scheduler_mainCoro(((IoState *)state)->scheduler);
	Coro_object_(DATA(self)->cid, self);
	Coro_unyieldCallback_(DATA(self)->cid, (CoroUnyieldCallback *)IoCoroutine_prepareToUnyield);

	return self;
}

void IoCoroutine_protoFinish(IoCoroutine *self)
{
	IoMethodTable methodTable[] = {
	{"ioStack", IoCoroutine_ioStack},
	{"yield", IoCoroutine_yield},
	{"pause", IoCoroutine_pause},
	{"resume", IoCoroutine_resume},
	{"scheduleNext", IoCoroutine_scheduleNext},
	{"run", IoCoroutine_run},
	{"setLabel", IoCoroutine_setLabel},
	{"label", IoCoroutine_label},
	{"main", IoCoroutine_main},
	{NULL, NULL},
	};
	
	IoObject_addMethodTable_(self, methodTable);
}

IoCoroutine *IoCoroutine_rawClone(IoCoroutine *proto) 
{ 
	IoObject *self = IoObject_rawClonePrimitive(proto);
	self->data = calloc(1, sizeof(IoCoroutineData));
	DATA(self)->ioStack = Stack_new();
	DATA(self)->cid = (Coro *)0x0;
	return self; 
}

IoCoroutine *IoCoroutine_new(void *state)
{
	IoObject *proto = IoState_protoWithInitFunction_((IoState *)state, IoCoroutine_proto);
	IoObject *self = IOCLONE(proto);
#ifdef DEBUG
	printf("Coroutine_%p new\n", self);
#endif
	IoCoroutine_rawSetLabel_(self, IOSYMBOL("unlabeled"));
	return self;
}

void IoCoroutine_free(IoCoroutine *self) 
{
	IoCoroutine_rawAbort(self);
	Stack_free(DATA(self)->ioStack);
	free(DATA(self));
}

void IoCoroutine_mark(IoCoroutine *self) 
{ 
	//printf("IoCoroutine_%p IoCoroutine_mark\n", (void *)self);
	//IoCoroutine_rawPrint(self);
	Stack_do_(DATA(self)->ioStack, (ListDoCallback *)IoObject_makeGrayIfWhite);
}

void IoCoroutine_makeGray(IoCoroutine *self) 
{
	/*
	printf("IoCoroutine_%p makeGray\n", (void *)self);
	if (!self || !ISCOROUTINE(self))
	{
		printf("IoCoroutine_makeGray error - self not coroutine\n");
		exit(-1);
	}
	*/

	IoState_makeGray_(IOSTATE, self);
	//IoObject_makeGrayIfWhite(self);
	//IoCoroutine_mark(self); // need to do this to cover stack pushes since mark
}

// raw

Stack *IoCoroutine_rawRetainedValues(IoCoroutine *self)
{ 
	return DATA(self)->ioStack; 
}

int IoCoroutine_rawRetainingCount(IoCoroutine *self)
{ 
	return Stack_totalSize(DATA(self)->ioStack); 
}

void IoCoroutine_rawShow(IoCoroutine *self)
{
	Stack_do_(DATA(self)->ioStack, (StackDoCallback *)IoObject_show);
	printf("\n");
}

void *IoCoroutine_cid(IoObject *self)
{
	return DATA(self)->cid;
}

// label

//#ifdef DEBUG

static const char *na = "n/a";

const char *IoCoroutine_labelCString(IoCoroutine *self)
{
	IoSymbol *label = IoCoroutine_rawLabel(self);
	return (label && ISSEQ(label)) ? CSTRING(label) : na;
}

void IoCoroutine_printLabel(IoCoroutine *self)
{
	printf("Coroutine_%p_%s", self, IoCoroutine_labelCString(self));
}

//#endif


void IoCoroutine_rawSetLabel_(IoCoroutine *self, IoSymbol *s)
{
	IoObject_setSlot_to_(self, IOSYMBOL("label"), s);
}

IoObject *IoCoroutine_rawLabel(IoCoroutine *self)
{
	return IoObject_rawGetSlot_(self, IOSYMBOL("label"));
}

IoObject *IoCoroutine_setLabel(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	IoCoroutine_rawSetLabel_(self, s);
	return self;
}

IoObject *IoCoroutine_label(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoObject *label = IoCoroutine_rawLabel(self);
	return label ? label : IONIL(self);
}

// runTarget

void IoCoroutine_rawSetRunTarget_(IoObject *self, IoObject *v)
{
	IoObject_setSlot_to_(self, IOSYMBOL("runTarget"), v);
}

IoObject *IoCoroutine_rawRunTarget(IoObject *self)
{
	return IoObject_rawGetSlot_(self, IOSYMBOL("runTarget"));
}

// runMessage

void IoCoroutine_rawSetRunMessage_(IoObject *self, IoObject *v)
{
	IoObject_setSlot_to_(self, IOSYMBOL("runMessage"), v);
	//printf("%p set runMessage = %s\n", (void *)self, CSTRING(IoMessage_name(v)));
}

IoObject *IoCoroutine_rawRunMessage(IoObject *self)
{
	return IoObject_rawGetSlot_(self, IOSYMBOL("runMessage"));
}

// runLocals

void IoCoroutine_rawSetRunLocals_(IoObject *self, IoObject *v)
{
	IoObject_setSlot_to_(self, IOSYMBOL("runLocals"), v);
}

IoObject *IoCoroutine_rawRunLocals(IoObject *self)
{
	return IoObject_rawGetSlot_(self, IOSYMBOL("runLocals"));
}

// parent

void IoCoroutine_rawSetParentCoroutine_(IoObject *self, IoObject *v)
{
	IoObject_setSlot_to_(self, IOSYMBOL("parentCoroutine"), v);
}

IoObject *IoCoroutine_rawParentCoroutine(IoObject *self)
{
	return IoObject_getSlot_(self, IOSYMBOL("parentCoroutine"));
}

// result

void IoCoroutine_rawSetResult_(IoObject *self, IoObject *v)
{
	IoObject_setSlot_to_(self, IOSYMBOL("result"), v);
}

IoObject *IoCoroutine_rawResult(IoObject *self)
{
	return IoObject_getSlot_(self, IOSYMBOL("result"));
}

// exception

void IoCoroutine_rawRemoveException(IoObject *self)
{
	IoObject_removeSlot_(self, IOSYMBOL("exception"));
}

void IoCoroutine_rawSetException_(IoObject *self, IoObject *v)
{
	IoObject_setSlot_to_(self, IOSYMBOL("exception"), v);
}

IoObject *IoCoroutine_rawException(IoObject *self)
{
	return IoObject_getSlot_(self, IOSYMBOL("exception"));
}

// ioStack

IoObject *IoCoroutine_ioStack(IoCoroutine *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("ioStack", "Returns List of values on this coroutine's stack.")
	*/
	
	return IoList_newWithList_(IOSTATE, Stack_asList(DATA(self)->ioStack));
}

// scheduler 

void IoCoroutine_rawReturnToParent(IoCoroutine *self)
{
	IoState *state = IOSTATE;
	IoCoroutine *parent = IoCoroutine_rawParentCoroutine(self);
	//Coro *thisCoro = DATA(self)->cid;
		  	
	if (parent && ISCOROUTINE(parent))
	{
		Coro *parentCid = DATA(parent)->cid;
		
		#ifdef DEBUG
		IoCoroutine_printLabel(self);
		printf(" returning ");
		IoCoroutine_printLabel(parent);
		printf("\n");
		#endif
		
		Coro_resume(parentCid);
		Scheduler_scheduleNext_(state->scheduler, parentCid);
		IoCoroutine_rawPause(self);
		//printf("IoCoroutine auto aborting coro %p\n", (void *)self);
	}
	else
	{
		if (self == IOSTATE->mainCoroutine)
		{
			printf("IoCoroutine error: attempt to return from main coro\n");
			exit(-1);
		}
	}

	if (!ISNIL(IoCoroutine_rawException(self)))
	{
		IoCoroutine_rawPrintBackTrace(self);
	}
	
	for (;;)
	{
		IoCoroutine_rawAbort(self);
		IoCoroutine_rawYield(self);
		printf("IoCoroutine error: unable to auto abort coro %p\n", (void *)self);
	}
}

void IoCoroutine_coroStart(void *context)
{
	IoObject *result;
	IoCoroutine *self = (IoCoroutine *)context;
	//void *cid = DATA(self)->cid;
	
#ifdef DEBUG
	IoCoroutine *currentCoroutine = IoState_currentCoroutine(IOSTATE);
	IoCoroutine_printLabel(currentCoroutine);
	printf(" starting ");
	IoCoroutine_printLabel(self);
	printf("\n");
#endif	

	IoState_setCurrentCoroutine_(IOSTATE, self);
     IoState_yield(IOSTATE);
	
	result = IoMessage_locals_performOn_(IOSTATE->mainMessage, self, self);
	
	IoCoroutine_rawSetResult_(self, result);
	IoCoroutine_rawReturnToParent(self);
}

IoObject *IoCoroutine_main(IoObject *self, IoObject *locals, IoMessage *m)
{
	IoObject *runTarget  = IoCoroutine_rawRunTarget(self);
	IoObject *runLocals  = IoCoroutine_rawRunLocals(self);
	IoObject *runMessage = IoCoroutine_rawRunMessage(self);
	
	if (runTarget && runLocals && runMessage)
	{
		return IoMessage_locals_performOn_(runMessage, runLocals, runTarget);
	}
	else
	{
		printf("IoCoroutine_main()  missing needed parameters\n");
	}

	return IONIL(self);
}

void IoCoroutine_rawRun(IoObject *self)
{
	IoState *state = IOSTATE;
	
	if (DATA(self)->cid)
	{
		printf("error: attempt to run same coro twice\n");
		exit(-1);
	}
	else
	{ 
		//IoCoroutine *current = IoState_currentCoroutine(state);
		Coro *coro = Scheduler_newCoro(state->scheduler);
		Coro_label_(coro, "IoCoroutine");
		Coro_object_(coro, self);
		Coro_unyieldCallback_(coro, (CoroUnyieldCallback *)IoCoroutine_prepareToUnyield);
		DATA(self)->cid = coro;
		//IoState_addCoroutine_(state, self);
		Scheduler_startCoro(state->scheduler, coro, (void *)self, IoCoroutine_coroStart);
		//printf("self %p coro %p tid %p\n", self, DATA(self), DATA(self)->cid);
	}
}

IoObject *IoCoroutine_run(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoCoroutine_rawRun(self);
	return IoCoroutine_rawResult(self);
}

IoCoroutine *IoCoroutine_newWithTry(void *state, 
							 IoObject *target, 
							 IoObject *locals, 
							 IoMessage *message)
{
	IoCoroutine *currentCoro = (IoCoroutine *)IoState_currentCoroutine((IoState *)state);
	IoCoroutine *self = IoCoroutine_new(state);
	IoCoroutine_rawSetLabel_(self, IOSYMBOL("newWithTry()"));
	IoCoroutine_rawSetRunTarget_(self, target);
	IoCoroutine_rawSetRunLocals_(self, locals);
	IoCoroutine_rawSetRunMessage_(self, message);
	//printf("Coroutine %p runMessage = %s\n", (void *)self, CSTRING(IoMessage_name(message)));
	IoCoroutine_rawSetParentCoroutine_(self, currentCoro);
	IoCoroutine_rawRun(self);
	//IoState_yield(state);
	IoCoroutine_rawPause(currentCoro); // pause the current one, not self, which is the new one
	#ifdef DEBUG
	//IoCoroutine_rawPrint(self);
	//printf(" after IoCoroutine_newWithTry\n");
	#endif
	//IoCoroutine_rawPause(currentCoro);
	//while (IoState_yieldIfNeeded(state)) {}
	return self;
}

void IoCoroutine_raiseError(IoCoroutine *self, IoSymbol *description, IoMessage *m)
{
	IoObject *e = IoObject_rawGetSlot_(self, IOSYMBOL("Exception"));
	
	if (e)
	{
		e = IOCLONE(e);
		IoObject_setSlot_to_(e, IOSYMBOL("error"), description);
		if (m) IoObject_setSlot_to_(e, IOSYMBOL("caughtMessage"), m);
		IoObject_setSlot_to_(e, IOSYMBOL("coroutine"), self);
		IoCoroutine_rawSetException_(self, e);
	}
	
	//printf("coro %p returning to parent coro. stack size = %i\n", (void *)self, Stack_count(DATA(self)->ioStack));
	
	IoCoroutine_rawReturnToParent(self);
}

// raw

void IoCoroutine_prepareToUnyield(IoObject *self)
{
	IoState_setCurrentCoroutine_(IOSTATE, self);
}

void IoCoroutine_rawYield(IoObject *self)
{
#ifdef DEBUG
	IoCoroutine_printLabel(self);
	printf(" yield\n");
#endif
	
	Scheduler_yield(IOSTATE->scheduler);
	
#ifdef DEBUG
	IoCoroutine_printLabel(self);
	printf(" unyield\n");
#endif
}

void IoCoroutine_rawAbort(IoObject *self)
{ 
	IoState *state = IOSTATE;
	Coro *cid = DATA(self)->cid;
	
	if (cid)
	{
		//IoState_removeCoroutine_(state, self);
		//printf("aborting %p\n", p);
		Scheduler_prepareToKillCoro_(state->scheduler, cid);
		DATA(self)->cid = 0x0;
	}
}

void IoCoroutine_rawPrepareToPause(IoObject *self)
{
	Coro *cid = DATA(self)->cid;
	
	if (cid)
	{
		Scheduler_prepareToPauseCoro_(IOSTATE->scheduler, cid);
	}
}

void IoCoroutine_rawPause(IoObject *self)
{
	Coro *cid = DATA(self)->cid;
	
	if (cid)
	{
		//printf("pausing %p\n", self);
		IoCoroutine_rawPrepareToPause(self);
		IoCoroutine_rawYield(self);
	}
}

unsigned int IoCoroutine_rawIsPaused(IoObject *self)
{ 
	IoState *state = IOSTATE;
	Coro *cid = DATA(self)->cid;
	
	return (cid && !Scheduler_hasCoro_(state->scheduler, cid));
}

void IoCoroutine_rawResume(IoObject *self)
{
	IoState *state = IOSTATE;
	Coro *cid = DATA(self)->cid;
	
	if (cid)
	{
		//printf("Object resuming %p\n", self);
		Scheduler_resumeCoro_(state->scheduler, cid);
	}
}

// methods

IoObject *IoCoroutine_yield(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoState_yield(IOSTATE);
	return self;
}

IoObject *IoCoroutine_resume(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoCoroutine_rawResume(self);
	return self; 
}

IoObject *IoCoroutine_pause(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoCoroutine_rawPause(self); 
	return self; 
}

IoObject *IoCoroutine_isActive(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	IoState *state = IOSTATE;
	
	if (!DATA(self) || !Scheduler_hasCoro_(state->scheduler, DATA(self)->cid)) 
	{ 
		return state->ioNil; 
	}
	
	return self;
}

IoObject *IoCoroutine_wait(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	double seconds = IoMessage_locals_doubleArgAt_(m, locals, 0);
	IoState *state = IOSTATE;
	IoObject *currentCoroutine = IoState_currentCoroutine(state);
	double waited = Scheduler_waitSeconds_(state->scheduler, seconds);
	IoState_setCurrentCoroutine_(state, currentCoroutine); // needed?
	return IONUMBER(waited);
}

IoObject *IoCoroutine_scheduleNext(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	Scheduler_scheduleNext_(IOSTATE->scheduler, DATA(self)->cid);
	return self;
}

// stack trace

int IoCoroutine_rawIoStackSize(IoObject *self)
{
	return Stack_count(DATA(self)->ioStack);
}

void IoCoroutine_rawPrint(IoObject *self)
{
	Coro *coro = DATA(self)->cid;
	
	if (coro)
	{
		printf("Coroutine_%p %s with cid %p '%s' stack %i\n", 
			  (void *)self, 
			  IoCoroutine_labelCString(self),
			  (void *)coro, 
			  Coro_label(coro),  
			  (int)Stack_count(DATA(self)->ioStack));
	}
}

// debugging

int IoCoroutine_rawDebuggingOn(IoObject *self)
{
	return DATA(self)->debuggingOn;
}

IoObject *IoCoroutine_turnOnMessageDebugging(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("turnOnMessageDebugging", 
		   "Turns on message level debugging for this coro. When on, this 
coro will send a vmWillSendMessage message to the Debugger object before 
each message send and pause itself. See the Debugger object documentation 
for more information. ")
	*/
	
	DATA(self)->debuggingOn = 1;
	IoState_updateDebuggingMode(IOSTATE);
	
	return self;
}

IoObject *IoCoroutine_turnOffMessageDebugging(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("turnOffMessageDebugging", "Turns off message level debugging for this coro. ")
	*/
	
	DATA(self)->debuggingOn = 0;
	IoState_updateDebuggingMode(IOSTATE);
	
	return self;
}

IoObject *IoObject_performWithDebugger(IoObject *self, IoObject *locals, IoMessage *m)
{
	IoState *state = IOSTATE;
	IoObject *currentCoroutine = IoState_currentCoroutine(state);
	
	if (IoCoroutine_rawDebuggingOn(currentCoroutine))
	{
		if (state->debugger)
		{
			/*
			 printf("IoObject_performWithDebugger(%i, %p, %s)\n", 
				   (int)(void *)self, (void *)locals, CSTRING(IoMessage_name(m)));
			 */
			IoMessage_setCachedArg_to_(state->vmWillSendMessage, 0, currentCoroutine);
			IoMessage_setCachedArg_to_(state->vmWillSendMessage, 1, m);
			IoMessage_setCachedArg_to_(state->vmWillSendMessage, 2, self);
			IoMessage_setCachedArg_to_(state->vmWillSendMessage, 3, locals);
			//IoObject_actorPerform(state->debugger, locals, state->vmWillSendMessage, 0);
			IoCoroutine_rawPause(currentCoroutine);
		}
	}
	
	return IoObject_perform(self, locals, m);
}

void IoCoroutine_rawPrintBackTrace(IoObject *self)
{
	IoObject *e = IoCoroutine_rawException(self);
	IoMessage *caughtMessage = IoObject_rawGetSlot_(e, IOSYMBOL("caughtMessage"));
	
	if (IoObject_rawGetSlot_(e, IOSYMBOL("showStack"))) // sanity check
	{
		IoState_on_doCString_withLabel_(IOSTATE, e, "showStack", "[Coroutine]");
	}
	else
	{
		IoSymbol *error = IoObject_rawGetSlot_(e, IOSYMBOL("error"));
		//IoMessage *caughtMessage = IoObject_rawGetSlot_(e, IOSYMBOL("caughtMessage"));
		
		if (error)
		{
			fputs(CSTRING(error), stderr); 
			fputs("\n", stderr); 
		}
		else
		{
			fputs("error: [missing error slot in Exception object]\n", stderr); 
		}
		
		if (caughtMessage)
		{
			ByteArray *ba = IoMessage_asMinimalStackEntryDescription(caughtMessage);
			fputs(ByteArray_asCString(ba), stderr); 
			fputs("\n", stderr); 
			ByteArray_free(ba);
		}
	}
}

