/*
 docCopyright("Steve Dekorte", 2002)
 docLicense("BSD revised")
 */

#include "IoMessage_parser.h"
#include "IoObject.h"
#include "IoSeq.h"
#include "IoNumber.h"
#include "IoState.h"
#include "IoLexer.h"
#include "IoToken_parser.h"
#include <ctype.h>

#define DATA(self) ((IoMessageData *)self->data)

void IoMessage_cacheIfPossible(IoMessage *self) 
{
	IoMessage *m = IoMessage_newFromText_label_(IOSTATE, 
																							CSTRING(DATA(self)->name), CSTRING(DATA(self)->label));
	IoMessage_cachedResult_(self, DATA(m)->cachedResult);
}

void IoMessage_ifPossibleCacheToken_(IoMessage *self, IoToken *p) 
{
	IoSymbol *method = DATA(self)->name;
    //printf("%s %i\n", CSTRING(method), p->token->type);
	
	switch ((int)IoToken_type(p))
	{
		case TRIQUOTE_TOKEN:
			IoMessage_cachedResult_(self, IoState_symbolWithCString_length_(IOSTATE, 
																																			IoSeq_asCString(method) + 3, 
																																			IoSeq_rawSize(method) - 6)); 
			break;
			
		case MONOQUOTE_TOKEN:
			IoMessage_cachedResult_(self, 
															IoSeq_rawAsUnescapedSymbol(IoSeq_rawAsUnquotedSymbol(method))); 
			break;
			
		case NUMBER_TOKEN:
			IoMessage_cachedResult_(self, 
															IONUMBER(IoSeq_asDouble(method)));
			break;
			
		case HEXNUMBER_TOKEN:
			IoMessage_cachedResult_(self, 
															IONUMBER(IoSeq_rawAsDoubleFromHex(method)));
			break;      
			
		default:
			if (IoSeq_rawEqualsCString_(method, "Nil"))
			{ 
				IoMessage_cachedResult_(self, IONIL(self)); 
			}
	}
}

IoMessage *IoMessage_newFromIoToken_(void *state, IoToken *p) 
{
	IoSymbol *s = IoState_symbolWithCString_((IoState *)state, IoToken_name(p));
	IoMessage *m = IoMessage_newWithName_(state, s);
	int i;
	
	IoMessage_ifPossibleCacheToken_(m, p);
	
	for (i = 0; i < List_size(p->args); i ++)
	{
		IoToken *parg = List_at_(p->args, i);
		IoMessage_addArg_(m, IoMessage_newFromIoToken_(state, parg));
	}
	
	if (p->attached) 
	{
		DATA(m)->attachedMessage = IoMessage_newFromIoToken_(state, p->attached);
	}
	
	if (p->next) 
	{
		DATA(m)->nextMessage = IoMessage_newFromIoToken_(state, p->next);
	}
	
	DATA(m)->lineNumber = IoToken_lineNumber(p);
	IoMessage_rawSetCharNumber_(m, IoToken_charNumber(p));
	
	return m;
}

IoMessage *IoMessage_newFromText_label_(void *state, const char *text, const char *label)
{
	IoMessage *m;
		
	IoState_pushRetainPool((IoState *)state);
	
	{
		IoToken *error = 0x0;
		IoSymbol *errorString = 0x0;
		IoSymbol *slabel = IoState_symbolWithCString_((IoState *)state, label);
		IoToken *p;
		IoLexer *lexer = IoLexer_new();
		
		IoLexer_string_(lexer, text);
		IoLexer_lex(lexer);
		
		/*
		 IoLexer_printTokens(lexer); 
		 printf("\n");
		 */
		
		if (IoLexer_errorToken(lexer))
		{ 
			m = IoMessage_newFromIoToken_(state, IoLexer_errorToken(lexer)); 
			IoMessage_label_(m, IoState_symbolWithCString_((IoState *)state, "[lexer]"));
			errorString = IoState_symbolWithCString_((IoState *)state, 
													IoToken_error(IoLexer_errorToken(lexer)));
		}
		else
		{
			p = IoLexer_rootMessageToken(lexer, &error);
			
			if (error) 
			{
				p = error;
			}
			
			if (!p) 
			{
				IoLexer_free(lexer);
				return IoMessage_newWithName_(state, IoState_symbolWithCString_((IoState *)state, "Nil"));
			}
			
			m = IoMessage_newFromIoToken_(state, p);
			
			/*IoMessage_printWithReturn(m);*/
			
			if (error) 
			{
				IoMessage_label_(m, IoState_symbolWithCString_((IoState *)state, "[parser]")); 
			} 
			else 
			{
				IoMessage_label_(m, slabel); 
			}
			
			if (error) 
			{
				errorString = IoState_symbolWithCString_((IoState *)state, error->error);
			}
		}
		
		IoLexer_free(lexer);
		
		if (errorString) 
		{
			//printf("IoMessage_newFromText_label_: %s\n\n", text);
			//printf("current coroutine %p\n", IoState_currentCoroutine(state)); 
			IoState_error_(state, m, "compile error: %s", CSTRING(errorString));
		}
		
		if (!(IoMessage_name(m)))
		{
			IoState_error_(state, 0x0, "internal compiler error: message with no method");
		}
		/*IoMessage_markTails(m);*/
	}
	
	IoState_popRetainPoolExceptFor_((IoState *)state, m);
	
	return m;
}

