/*   
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#define IOOBJECTGROUP_C  
#include "IoObjectGroup.h"
#undef IOOBJECTGROUP_C 
#include "IoState.h"
#include "IoObject.h"
#include "IoWeakLink.h"
#include "IoState_collector.h"

IoObjectGroup *IoObjectGroup_newWithColor_(void *state, unsigned char color)
{
    IoObjectGroup *self = (IoObjectGroup *)calloc(1, sizeof(IoObjectGroup));
    self->mark.color = color;
    self->state = state;
    return self;
}

void IoObjectGroup_free(IoObjectGroup *self) 
{ 
    free(self); 
}

/*
void IoObjectGroup_colorValues(IoObjectGroup *self)
{
    IoObject *v = self->first;
    unsigned char color = self->mark.color;
    
    while (v) 
    { 
        v->mark.color = color; 
        v = v->mark.next; 
    }
}
*/

size_t IoObjectGroup_count(IoObjectGroup *self)
{
    size_t count = 0;
    IoObject *v = self->first;
    
    while (v) 
    { 
        count ++; 
        v = v->mark.next; 
    }
    
    return count;
}

#include "Hash.h"

void IoObjectGroup_findLoops(IoObjectGroup *self)
{
    IoObject *v = self->first;
    IoObject *next;
    Hash *hash = Hash_new();
    
    while (v)
    { 
        if (Hash_at_(hash, v))
        {
            printf("\nIoObjectGroup_findLoops() error: loop found in IoObjectGroup!\n");
            exit(-1);
        }
        
        Hash_at_put_(hash, v, v);
        
        next = v->mark.next;
        v = next; 
    }
    
    Hash_free(hash);
}

/*
size_t IoObjectGroup_occurencesOf_(IoObjectGroup *self, IoObject *value)
{
	IoObject *v = self->first;
	IoObject *next;
	size_t count = 0;
	
	while (v)
	{ 
		if (v == value)
		{
			count ++;
		}
		
		v = v->mark.next;
	}
	
	return count;
}
*/

void IoObjectGroup_markValues(IoObjectGroup *self)
{
    IoObject *v = self->first;
    IoObject *next;
    
    //IoObjectGroup_findLoops(self);
    
    while (v)
    { 
        next = v->mark.next;
        IoObject_mark(v); 
        v = next; 
    }
    
    //printf("IoObjectGroup_markValues done\n");
}

size_t IoObjectGroup_freeValues(IoObjectGroup *self)
{
    size_t count = 0;
    IoObject *v = self->first;
    IoObject *next;
    
    while (v) 
    {
        count ++;
        next = v->mark.next;
        
        if (v->mark.hasWeakLink) 
        {
            IoState_willFreeWeakLinkedValue_(self->state, v);
        }
        
        if (v->persistentId) 
        {
            IoState_willFreePersistentValue_(self->state, v);
        }
        
        IoObject_free(v); 
        v = next;
    }
    
    self->first = 0x0;
    
    return count;
}

unsigned char IoObjectGroup_isEmpty(IoObjectGroup *self)
{ 
    return (self->first == 0x0); 
}

int IoObjectGroup_sharesWithGroup_(IoObjectGroup *self, IoObjectGroup *other)
{
    IoObject *v = self->first;
    
    while (v)
    { 
        if (IoObjectGroup_hasValue_(other, v)) 
        {
            return 1; 
        }
        
        v = v->mark.next; 
    }
    
    return 0;
}

int IoObjectGroup_hasValue_(IoObjectGroup *self, IoObject *value)
{
    IoObject *v = self->first;
    
    while (v)
    { 
        if (v == value) 
        {
            return 1;
        }
        
        v = v->mark.next; 
    }
    
    return 0;
}

void IoObjectGroup_show(IoObjectGroup *self)
{
    IoObject *v = self->first;
    
    while (v)
    {
        IoObject_show(v);
        v = v->mark.next;
    } 
}
