/*
 docCopyright("Steve Dekorte", 2002)
 docLicense("BSD revised")
 */

#include "IoState.h"

#include "IoObject.h"
#include "IoObjectGroup.h"
#include "IoCoroutine.h"
#include "IoSeq.h"
#include "IoNumber.h"
#include "IoWeakLink.h"

#include <time.h>
#include <limits.h> 
#include <stdlib.h>

void IoState_updateCachedCollectorColors(IoState *self)
{
	self->currentBlackColor = self->blackGroup->mark.color;
	self->currentWhiteColor = self->whiteGroup->mark.color;
}

IoObject *IoState_retain_(IoState *self, IoObject *v)
{ 
	return (IoObject *)List_append_(self->retainedValues, v); 
}

void IoState_release_(IoState *self, IoObject *v)
{ 
	List_removeLast_(self->retainedValues, v); 
}

void IoState_gcInitPhase(IoState *self)
{
	// assume all values currently white 
	
	if (self->lobby)
	{
		IoObject_makeGrayIfWhite(self->lobby);
		IoObject_mark(self->lobby);
	}
}

/*
 void IoState_autoAdaptParameters(IoState *self)
 {
	 // dynamically adjust wait until sweep size 
	 
	 self->lastReferencedObjectCount = IoObjectGroup_count(self->blackGroup);
	 self->allocsPerSweep  = 
		 self->lastReferencedObjectCount > STATE_DEFAULT_COLECT_AFTER ? 
		 self->lastReferencedObjectCount : STATE_DEFAULT_COLECT_AFTER;
	 
	 //printf("self->allocsPerSweep = %i\n", self->allocsPerSweep);	
 }
 
 void IoState_testForSharing(IoState *self)
 {
	 int sharing = 0;

	 if(IoObjectGroup_sharesWithGroup_(self->whiteGroup, self->grayGroup))
	  { 
		  printf("IoObjectGroup_sharesWithGroup_(self->whiteGroup, self->grayGroup)\n"); 
		  sharing = 1;
	  }
	  
	  if(IoObjectGroup_sharesWithGroup_(self->whiteGroup, self->blackGroup))
	  { 
		  printf("IoObjectGroup_sharesWithGroup_(self->whiteGroup, self->blackGroup)\n"); 
		  sharing = 1;
	  }
	  
	  if(IoObjectGroup_sharesWithGroup_(self->grayGroup, self->blackGroup))
	  { 
		  printf("IoObjectGroup_sharesWithGroup_(self->grayGroup, self->blackGroup)\n"); 
		  sharing = 1;
	  }
	  
	  if (sharing)
	  {
		  exit(-1);
	  }
}
 */

int IoState_gcSweepPhase(IoState *self)
{
	//int count, debug = 1; 
	int count, debug = self->debugCollector;
	
	// mark the permanentally retained values
	
	if (debug) 
	{
		fputs("--- collector start ---\n", stdout);
		printf("protos: %i ", (int)Hash_count(self->primitiveProtos));
		printf("perm: %i\n", (int)List_size(self->retainedValues));
	}
	
	List_do_(self->retainedValues, (ListDoCallback *)IoObject_makeGrayIfWhite);
	
	// Marking the active coros will mark their stacks
	// non-active coros will be reached by reference 
	
	Scheduler_doOnObjects_(self->scheduler, (SchedulerDoCallback *)IoCoroutine_makeGray);
	IoCoroutine_makeGray(self->mainCoroutine);
	IoCoroutine_makeGray(self->currentCoroutine);
	
	Hash_do_(self->primitiveProtos, (PHashDoCallback *)IoObject_makeGrayIfWhite);
	
	if (debug)
	{		
		printf("black: %i ", (int)IoObjectGroup_count(self->blackGroup));
		printf("gray:  %i ", (int)IoObjectGroup_count(self->grayGroup));
		printf("white: %i \n", (int)IoObjectGroup_count(self->whiteGroup));
		printf("recycled: %i\n", (int)List_size(self->recycledObjects));
		
		//IoObjectGroup_findLoops(self->whiteGroup);
		//IoObjectGroup_findLoops(self->grayGroup);
		//IoObjectGroup_findLoops(self->blackGroup);
		//IoState_testForSharing(self);
	}
	
	while (!IoObjectGroup_isEmpty(self->grayGroup)) 
	{
		IoObjectGroup_markValues(self->grayGroup);
	}
	
	if (debug) 
	{
		printf("left:  %i\n", (int)IoObjectGroup_count(self->blackGroup));
	}
	
	// free the white group - no need to have a free group 
	count = IoObjectGroup_freeValues(self->whiteGroup);
	
	// swap black group with the now empty white group 	
	self->sweepsSinceLastGeneration ++;

	
	if (!self->sweepsPerGeneration ||
	    self->sweepsSinceLastGeneration == self->sweepsPerGeneration)
	{
		IoObjectGroup *vg = self->blackGroup;
		
		self->blackGroup = self->whiteGroup;
		self->whiteGroup = vg;
		self->sweepsSinceLastGeneration = 0;
		
		IoState_updateCachedCollectorColors(self);
	}
	
	
	IoState_gcInitPhase(self);
	self->newValueSinceSweepCount = 0;
	
	if (debug)
	{
		//printf("generation: %i\n", self->sweepsSinceLastGeneration);
		printf("freed: %i\n", (int)count);
		printf("symbols: %i\n", (int)SkipDB_count(self->symbols));
		printf("coros: %i\n", (int)Scheduler_coroCount(self->scheduler));
		Scheduler_doOnObjects_(self->scheduler, (SchedulerDoCallback *)IoCoroutine_rawPrint); 
		fputs("--- collector done  ---\n", stdout);
	}
	
	return count;
}

int IoState_collectGarbage(IoState *self) 
{ 
	return IoState_gcSweepPhase(self); 
}

void IoState_makeGray_(IoState *self, IoObject *v)
{
	unsigned int c = v->mark.color; 
	unsigned int wc = self->whiteGroup->mark.color;
	unsigned int gc = self->grayGroup->mark.color;
	//unsigned int bc = self->blackGroup->mark.color;

	if (c != gc)
	{
		if (c == wc)
		{
			IoObjectGroup_isRemoving_(self->whiteGroup, v);
		}
		else
		{
			IoObjectGroup_isRemoving_(self->blackGroup, v);
		}
		
		IoObjectGroup_append_(self->grayGroup, v); 
	}
}

void IoState_makeBlackFromGray_(IoState *self, IoObject *v)
{ 
	IoObjectGroup_isRemoving_(self->grayGroup, v);
	IoObjectGroup_append_(self->blackGroup, v); 
}

void IoState_makeGrayFromWhite_(IoState *self, IoObject *v)
{ 
	/*
	if (self->grayGroup->first == v)
	{
		printf("error - already in gray group\n");
	}
	*/
	
	IoObjectGroup_isRemoving_(self->whiteGroup, v);
	IoObjectGroup_append_(self->grayGroup, v); 
}

void IoState_makeGrayFromBlack_(IoState *self, IoObject *v)
{ 
	IoObjectGroup_isRemoving_(self->blackGroup, v);
	IoObjectGroup_append_(self->grayGroup, v); 
}

void IoState_immediatelyFreeIfUnreferenced_(IoState *self, IoObject *v)
{
#ifdef IO_BLOCK_RECYCLING
	if (0x0 == IoObject_isReferenced(v))
	{ 
		IoState_immediatelyFree_(self, v); 
	}
#endif
}

void IoState_immediatelyFree_(IoState *self, IoObject *v)
{ 
	unsigned int color = v->mark.color;
	
	if (color == self->whiteGroup->mark.color)
	{ 
		IoObjectGroup_isRemoving_(self->whiteGroup, v); 
		/*IoObjectGroup_append_(self->freeGroup, v);*/
	}
	else if (color == self->grayGroup->mark.color)
	{ 
		IoObjectGroup_isRemoving_(self->grayGroup, v); 
	}
	else if (color == self->blackGroup->mark.color)
	{ 
		IoObjectGroup_isRemoving_(self->blackGroup, v); 
	}
	
	IoObject_free(v);
	
	/*if (self->newValueCount) */
	self->newValueCount --;
	/*printf("unref newValueCount = %i\n", self->newValueCount);*/
}

/*
 void IoState_immediatelyFreeIfUnreferenced_(IoState *self, IoObject *v)
 { 
	 if (0 == v->mark.refCount) IoState_immediatelyFree_(self, v);
 }
 */

void IoState_gcMarkPhase(IoState *self)
{
	IoObjectGroup_markValues(self->grayGroup);
	
	self->newValueSinceSweepCount += self->newValueCount;
	self->newValueCount = 0;
	
	//if ( IoObjectGroup_isEmpty(self->grayGroup) ||
	if (self->newValueSinceSweepCount > self->allocsPerSweep) 
	{
		 if (Stack_count(self->currentIoStack) > 40000)
		 { 
			 IoState_error_(self, 0x0, "stack overflow (about %i pointers)", 
						 Stack_count(self->currentIoStack)/sizeof(char *)); 
		 }
		
		IoState_gcSweepPhase(self);
		self->newValueSinceSweepCount = 0;
	}
}

int IoState_hasValue_(IoState *self, IoObject *v)
{ 
	return IoObjectGroup_hasValue_(self->whiteGroup, v) || 
	IoObjectGroup_hasValue_(self->grayGroup,  v) ||
	IoObjectGroup_hasValue_(self->blackGroup, v);
}

int IoState_garbageCollectorIsPaused(IoState *self)
{ 
	return (self->collectorPauseCount != 0); 
}

void IoState_pushCollectorPause(IoState *self)
{
	self->collectorPauseCount ++;
}

void IoState_popCollectorPause(IoState *self)
{
	self->collectorPauseCount --;
	
	if ((self->collectorPauseCount == 0) && 
	    (self->newValueCount > self->newValueCountToUpdate)) 
	{
		IoState_gcMarkPhase(self);
	}
	else if (self->collectorPauseCount < 0)
	{
		IoState_error_(self, 0x0, "negative collector pause count");     
	}
}

/*
void IoState_testAddValue_(IoState *self, IoObject *v)
{ 
	 if (!(v->tag))
	 {
		 IoState_fatalError_(self, "IoState_addValue_() Error: missing tag\n");
	 }
	
	//printf("IoState_addValue_(%p)\n", (void *)v);
	//printf("newValueCount = %i\n", self->newValueCount);
	
	 if ((self->collectorPauseCount == 0) && 
		(self->newValueCount > self->newValueCountToUpdate)) 
	 {
		 IoState_gcMarkPhase(self);
	 }
	 
	 if (IoState_hasValue_(self, v))
	 {
		 IoState_fatalError_(self, "error - value added more than once to IoState");
	 }
}
*/

IoObject *IoState_addValue_(IoState *self, IoObject *v)
{ 
	//IoState_testAddValue_(self, v);

	//IoObjectGroup_findLoops(self->whiteGroup);
	//IoObjectGroup_findLoops(self->grayGroup);
	//IoObjectGroup_findLoops(self->blackGroup);
			
	if (!v->mark.knownToCollector)
	{	
	/*	
		if (v->mark.next)
		{
			printf("collector error - v->mark.next\n");
		}
		
		if (v->mark.previous)
		{
			printf("collector error - v->mark.previous\n");
		}
		*/
		
		/*
		if (self->grayGroup->first == v)
		{
			int wcount = IoObjectGroup_occurencesOf_(self->whiteGroup, v);
			int gcount = IoObjectGroup_occurencesOf_(self->grayGroup, v);
			int bcount = IoObjectGroup_occurencesOf_(self->blackGroup, v);
			printf("color = %s\n", IoObject_markColorName(v));
			printf("collector error - attempt to add gray to white group\n");
		}
		*/

		v->mark.knownToCollector = 1;
		
		IoObjectGroup_append_(self->whiteGroup, v);
		//IoObjectGroup_findLoops(self->whiteGroup);
		self->newValueCount ++;
	}
	/*
	else
	{
		printf("object known to collector\n");
	}
	*/
	
	IoState_stackRetain_(self, v);
	
	//IoObjectGroup_findLoops(self->whiteGroup);
	//IoObjectGroup_findLoops(self->grayGroup);
	//IoObjectGroup_findLoops(self->blackGroup);

	//printf("new %s\n", v->tag->name);
	return v;
}

// weak links ---------------------------------------------

void IoState_addWeakLink_(IoState *self, IoWeakLink *w)
{
	IoObject *link = IoWeakLink_rawLink(w);
	IoWeakLink *first = Hash_at_(self->valueToWeakLink, link);
	
	if (first)
	{
		// not the first, so add it to the chain for this value
		IoWeakLink_appendWeakLink_(first, w);
	}
	else
	{
		// its the first, so set it in the hash 
		// printf("adding value %p with weaklink %p\n", link, w);
		Hash_at_put_(self->valueToWeakLink, link, w);
	}
}

void IoState_removeWeakLink_(IoState *self, IoWeakLink *w)
{
	// the WeakLink will unlink itself, we just need to check if its the first 
	IoWeakLink *first = Hash_at_(self->valueToWeakLink, IoWeakLink_rawLink(w));
	
	if (first == w)
	{
		IoObject *k = IoWeakLink_rawLink(first);
		IoObject *v = IoWeakLink_rawNextWeakLink(first);
		Hash_at_put_(self->valueToWeakLink, k, v);
	}
}

void IoState_willFreeWeakLinkedValue_(IoState *self, IoObject *value)
{
	IoWeakLink *first = Hash_at_(self->valueToWeakLink, value);
	//printf("collecting value %p with first weaklink %p\n", value, first);
	
	if (first) 
	{
		IoWeakLink_collected(first);
		//Stack_push_(self->collectedWeakLinks, first);
		Hash_at_put_(self->valueToWeakLink, value, 0x0);
	}
}

/* --- persistence -------------------------------- */

void IoState_willFreePersistentValue_(IoState *self, IoObject *value)
{
	if (!self->isCleaningUp)
	{ 
		IoStore_willFreePersistentObject_(self->store, value);
	}
}

// ------------------------------------------------ 

size_t IoState_memorySize(IoState *self)
{ 
	size_t total = 0;
	/*
	 total += IoObjectGroup_sum_(self->whiteGroup, IoObject_memorySize);
	 total += IoObjectGroup_sum_(self->grayGroup,  IoObject_memorySize);
	 total += IoObjectGroup_sum_(self->blackGroup, IoObject_memorySize);
	 */
	return total;
}

void IoState_compact(IoState *self)
{ 
	/*
	 IoObjectGroup_do_(self->whiteGroup, IoObject_compact);
	 IoObjectGroup_do_(self->grayGroup,  IoObject_compact);
	 IoObjectGroup_do_(self->blackGroup, IoObject_compact);
	 */
}
