/*#io
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#define QUEUE_C
#include "Queue.h"
#undef QUEUE_C
#include <stdlib.h>

Queue *Queue_new(void)
{
    Queue *self = (Queue *)calloc(1, sizeof(Queue));
    self->size = QUEUE_START_SIZE;
    self->items = (void **)calloc(1, self->size * sizeof(void *));
    return self;
}

Queue *Queue_clone(Queue *self)
{
    Queue *child = (Queue *)cpalloc(self, sizeof(Queue));
    child->items = (void **)cpalloc(self->items, self->size * sizeof(void *));
    return child;
}

void Queue_free(Queue *self)
{
    free(self->items);
    free(self);
}

int Queue_length(Queue *self) 
{ 
    return self->bottom - self->top; 
}

void *Queue_at_(Queue *self, int i)
{
    if (i < 0 || i >= Queue_length(self)) 
    {
		return (void *)0x0;
    }
    
    return self->items[self->top + i];
}

void Queue_compact(Queue *self)
{
    if (self->top)
    {
		memmove(self->items, self->items + self->top, self->bottom - self->top);
		self->bottom -= self->top;
		self->top = 0;
		self->size = self->bottom + 1;
		self->items = (void **)realloc(self->items, self->size);
    }
}

size_t Queue_memorySize(Queue *self)
{ 
    return sizeof(Queue) + (self->size * sizeof(void *)); 
}

void Queue_print(Queue *self)
{
    int i;

	/*
	 printf("Queue: \n");
	 printf("size   = %i\n", self->size);
	 printf("length = %i\n", Queue_length(self));
	 printf("top    = %i\n", self->top);
	 printf("bottom = %i\n", self->bottom);
	 */
	 	
    for (i = 0; i < self->size; i ++)
    { 
		void *p = self->items[i];
		
		if (i >= self->top && i < self->bottom)
		{
			printf(" %i : %p\n", i, (void *)p); 
		}
		else
		{
			printf(" - : %p\n",  (void *)p); 
		}
		/*
		 if (p == 0x0)
		 {
			 printf("-\n"); 
		 }
		 else
		 {
			 printf("%p\n", p); 
		 }
		 */	 
		
    }
    
    printf("\n");
}

void Queue_do_(Queue *self, QueueDoCallback *callback)
{
    int i, max = self->bottom;
    
    for (i = self->top; i < max; i ++) 
    {
		(*callback)(self->items[i]);
    }
}

void *Queue_firstTrueFor_with_(Queue *self, QueueFirstTrueForWithCallback *callback, void *arg)
{
    int i, max = self->bottom;
    
    for (i = self->top; i < max; i ++)
    {
		void *item = self->items[i];
		
		if ((*callback)(item, arg)) 
		{
			return item;
		}
    }
    
    return 0x0;
}

/*


#include "List.h"

int main(void)
{
    Queue *q = Queue_new();
    List *l = List_new();
    int i;
    
    Queue_print(q);
    
    for (i = 0; i < 1000; i ++)
    {
		if (random() % 2 == 0)
		{
			void *p1 = List_pop(l);
			void *p2 = Queue_pop(q);
			
			printf("pop list: %i queue: %i\n", (int)p1, (int)p2);
			Queue_print(q);
			
			if (p1 != p2)
			{
				printf("error\n");
				exit(1);
			}
		} 
		else 
		{
			void *p = (void *)((i % 15) + 1);
			List_at_insert_(l, 0, p);
			Queue_push_(q, p);
			printf("push %i\n", (int)p);
			Queue_print(q);
		}
    }
    
    List_free(l);
    Queue_free(q);
}
*/

/*
 Queue_push_(q, (void *)0x1); Queue_print(q);
 Queue_push_(q, (void *)0x2); Queue_print(q);
 printf("push %p\n", Queue_pop(q)); Queue_print(q);
 
 
 printf("pop %p\n", Queue_pop(q)); Queue_print(q);
 printf("pop %p\n", Queue_pop(q)); Queue_print(q);
 printf("pop %p\n", Queue_pop(q)); Queue_print(q);
 
 Queue_push_(q, (void *)0x3); Queue_print(q);
 Queue_push_(q, (void *)0x4); Queue_print(q);
 Queue_push_(q, (void *)0x5); Queue_print(q);
 
 printf("pop %p\n", Queue_pop(q)); Queue_print(q);
 printf("pop %p\n", Queue_pop(q)); Queue_print(q);
 printf("pop %p\n", Queue_pop(q)); Queue_print(q);

#endif
 */

