/*   
 *   copyright: Steve Dekorte, 2002. All rights reserved.
 *   license: See _BSDLicense.txt.
 */

#include "Base.h" 

#define Data_C
#include "Data.h"
#undef Data_C

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

Data *Data_new(void)
{
    Data *self = (Data *)calloc(1, sizeof(Data));
    Data_retain(self);
    return self;
}

void Data_free(Data *self)
{
    free(self->bytes);
    free(self);
}

Data *Data_newWithBytes_size_copy_(uint8_t *bytes, size_t byteCount, int copy)
{
    Data *self = Data_new();
    
    if (copy)
    {
	self->bytes = realloc(self->bytes, byteCount);
	memmove(self->bytes, bytes, byteCount);
    }
    else
    {
	self->bytes = data;
    }
    
    return self;
}

Data *Data_clone(Data *self)
{
    return Data_newWithBytes_size_copy_(self->bytes, self->size, 1);
}

/* --- reference counting --- */

void Data_retain(Data *self)
{
    self->refCount ++;
}

void Data_release(Data *self)
{
    self->refCount --;
    
    if (self->refCount == 0)
    {
	Data_free(self);
    }
}

/* --- size --- */

void Data_setSize_(Data *self, size_t newMemSize)
{
    if (newMemSize > self->memSize)
    {
	self->bytes = realloc(self->bytes, newMemSize);
	self->memSize = newMemSize;
    }
    
    self->size = newMemSize;
}

size_t Data_size(Data *self)
{
    return self->size;
}

/* --- mutation --- */

void Data_setBytes_size_(Data *self, uint8_t *bytes, size_t size)
{
    Data_setSize_(self, size);
    memcpy(self->bytes, bytes, size);
}

void Data_removeFrom_size_(Data *self, size_t start, size_t size)
{
    size_t end = start + size;
    
    if (end <= self->size)
    {
	size_t tail = self->size - end;
	
	if (tail)
	{
	    memcpy(self->bytes + start, self->bytes + end, tail);
	}
	
	Data_setSize_(self, end);
    }
}

inline uint8_t *Data_bytesAt_(Data *self, size_t index)
{
    return self->bytes + index;
}

void Data_at_insertBytes_size_(Data *self, size_t index, uint8_t *bytes, size_t size)
{
    size_t oldSize = self->size;
    size_t newSize = oldSize + size;
    size_t tailSize = newSize - oldSize;
    uint8_t *insertStart = Data_bytesAt_(self, index);
    uint8_t *insertEnd = Data_bytesAt_(self, index + size);
    
    Data_setSize_(self, newSize);

    /* move data to right to make room for insert */
    
    memmove(insertEnd, insertStart, tailSize);
    memcpy(insertStart, bytes, size);
}

/* --- access --- */

uint8_t *Data_bytesAt_size_(Data *self, size_t index, size_t size)
{
    if (index + size <- self->size)
    {
	/* assume memory transaction has begun */
	return self->bytes + index;
    }
    
    return 0x0;
}

void Data_endTransaction(Data *self)
{
    return;
}

