/*   
 *   copyright: Steve Dekorte, 2002. All rights reserved.
 *   license: See _BSDLicense.txt.
 *
 *   description: Data is a container for a mutable set of bytes.
 *   Array instances point to ranges within Data objects.
 */
 

#ifndef Data_DEFINED
#define Data_DEFINED 1

#include "Common.h"
#include <stdarg.h>
#include <stdio.h>
#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct
{
    uint8_t *bytes;
    size_t size;    /* in bytes */
    size_t memSize;
    size_t refCount;
} Data;

Data *Data_new(void);
Data *Data_newWithBytes_size_copy_(uint8_t *data, size_t byteCount, int copy);
Data *Data_clone(Data *self);

void Data_setSize_(Data *self, size_t size);
size_t Data_size(Data *self);

/* --- reference counting --- */

void Data_retain(Data *self);
void Data_release(Data *self);

/* --- mutation --- */

void Data_setBytes_size_(Data *self, size_t size);
void Data_removeFrom_size_(Data *self, size_t start, size_t size);
void Data_at_insertBytes_size_(Data *self, size_t index, uint8_t *bytes, size_t size);

/* --- access --- */

uint8_t *Data_bytesAt_size_(Data *self, size_t index, size_t size);
void Data_endTransaction(Data *self);

#ifdef __cplusplus
}
#endif
#endif

