/*   
 *   copyright: Steve Dekorte, 2002. All rights reserved.
 *   license: See _BSDLicense.txt.
 *
 * MarkedStack - array of void pointers
 * supports setting marks - when a mark is popped,
 * all stack items above it are popped as well 
 */

#ifndef MARKEDSTACK_DEFINED
#define MARKEDSTACK_DEFINED 1

#include "Common.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>

#ifdef LOW_MEMORY_SYSTEM
  #define MARKEDSTACK_START_SIZE 512
  #define MARKEDSTACK_RESIZE_FACTOR 2
#else
  #define MARKEDSTACK_START_SIZE 512
  #define MARKEDSTACK_RESIZE_FACTOR 2
#endif

typedef void (MarkedStackDoCallback)(void *);
typedef void (MarkedStackDoOnCallback)(void *, void *);

typedef struct
{
  void *ptr;
  unsigned int mark;
} StackItem;

typedef struct
{
  StackItem *items;
  int count;
  int maxCount;
} MarkedStack;

MarkedStack *MarkedStack_new(void);
void MarkedStack_free(MarkedStack *self);
void MarkedStack_clear(MarkedStack *self);
MarkedStack *MarkedStack_clone(MarkedStack *self);
void MarkedStack_copy_(MarkedStack *self, MarkedStack *other);

size_t MarkedStack_memorySize(MarkedStack *self);
void MarkedStack_compact(MarkedStack *self);

void MarkedStack_resize(MarkedStack *self);
size_t MarkedStack_totalSize(MarkedStack *self);

void MarkedStack_popToMark_(MarkedStack *self, ptrdiff_t mark);

/* not high performance */
void MarkedStack_do_on_(MarkedStack *self, MarkedStackDoOnCallback *callback, void *target);
MarkedStack *MarkedStack_newCopyWithoutMarks(MarkedStack *self);

#include "MarkedStack_inline.h"

#endif
