/*   copyright: Steve Dekorte, 2002
 *   All rights reserved. See _BSDLicense.txt.
 *
 * MarkedStack - array of void pointers
 * supports setting marks - when a mark is popped,
 * all stack items above it are popped as well 
 */

#ifdef MARKEDSTACK_C 
#define IO_IN_C_FILE
#endif
#include "Common_inline.h"
#ifdef IO_DECLARE_INLINES

IOINLINE int MarkedStack_count(MarkedStack *self) 
{ return (self->top - self->items); }

IOINLINE void MarkedStack_push_(MarkedStack *self, void *item)
{
  self->top ++;
  if (self->top == self->memEnd) MarkedStack_resize(self);
  *(self->top) = item;
}

IOINLINE ptrdiff_t MarkedStack_pushMark(MarkedStack *self)
{
  MarkedStack_push_(self, (void *)self->lastMark);
  self->lastMark = self->top - self->items;
  return self->lastMark;
}

IOINLINE void *MarkedStack_pop(MarkedStack *self)
{
  void *top = *(self->top);
  if (self->items != self->top) self->top --;
  return top;
} 

IOINLINE void MarkedStack_popMark(MarkedStack *self)
{
  self->top = self->items + self->lastMark - 1;
  if (self->lastMark) self->lastMark = (ptrdiff_t)(self->items[self->lastMark]);
}

IOINLINE void *MarkedStack_top(MarkedStack *self) { return *(self->top); }

IOINLINE void *MarkedStack_at_(MarkedStack *self, int i)
{ return self->items[i + 1]; } 

IOINLINE void MarkedStack_do_(MarkedStack *self, MarkedStackDoCallback *callback)
{
  void **itemP = self->top;
  ptrdiff_t mark = self->lastMark;
  while (itemP > self->items)
  { 
    if (itemP - self->items == mark) 
    { mark = (ptrdiff_t)(*itemP); }
    else 
    { (*callback)(*itemP); }
    itemP --;
  }
}

#undef IO_IN_C_FILE
#endif

