/*   
 *   copyright: Steve Dekorte, 2002. All rights reserved.
 *   license: See _BSDLicense.txt.
 *
 * Stack - array of void pointers
 * supports setting marks - when a mark is popped,
 * all stack items above it are popped as well 
 *
 * Designed to optimize push, pushMark and popMark 
 * at the expense of pop (since pop requires a mark check)
 */

#ifndef STACK_DEFINED
#define STACK_DEFINED 1

#include "Common.h"
#include "List.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>

typedef void (StackDoCallback)(void *);
typedef void (StackDoOnCallback)(void *, void *);

typedef List Stack;

Stack *Stack_new(void);
void Stack_free(Stack *self);
void Stack_clear(Stack *self);
Stack *Stack_clone(Stack *self);
void Stack_copy_(Stack *self, Stack *other);

size_t Stack_memorySize(Stack *self);
void Stack_compact(Stack *self);

void Stack_resize(Stack *self);
size_t Stack_totalSize(Stack *self);

void Stack_popToMark_(Stack *self, ptrdiff_t mark);

/* not high performance */
void Stack_do_on_(Stack *self, StackDoOnCallback *callback, void *target);
Stack *Stack_newCopyWithoutMarks(Stack *self);

#include "Stack_inline.h"

#endif
