/*   copyright: Steve Dekorte, 2002
 *   All rights reserved. See _BSDLicense.txt.
 *
 * Stack - array of void pointers
 * supports setting marks - when a mark is popped,
 * all stack items above it are popped as well 
 */

#ifdef STACK_C 
#define IO_IN_C_FILE
#endif
#include "Common_inline.h"
#ifdef IO_DECLARE_INLINES

IOINLINE int Stack_count(Stack *self) 
{ return List_count(self); }

IOINLINE void Stack_push_(Stack *self, void *item)
{ List_push_(self, item); }

IOINLINE void Stack_pushMark(Stack *self)
{
  List_push_(self, 0x0);
}

IOINLINE ptrdiff_t Stack_pushMarkPoint(Stack *self)
{
  List_push_(self, 0x0);
  return List_count(self);
}

IOINLINE void *Stack_pop(Stack *self)
{ return List_pop(self); } 

IOINLINE void Stack_popMark(Stack *self)
{
  int count = List_count(self);
  if (count)
  {
      int i = count - 1;
      if (i < 1) return;
      while (i)
      {
	if (self->items[i] == 0x0)
	{ 
	  List_setLength_(self, i);
	  return;
	}
	i --;
      }
  }
}

IOINLINE void *Stack_top(Stack *self) 
{ return List_top(self); }

IOINLINE void *Stack_at_(Stack *self, int i)
{ return List_at_(self, i); } 

IOINLINE void Stack_do_(Stack *self, StackDoCallback *callback)
{ List_do_(self, callback); }

#undef IO_IN_C_FILE
#endif

